angular.module("beamng.apps")
/**
 * Provides colour classes and some additional colour related functions
 */
.factory("TachoApps", ["TachoColours", "TachoBase", "TachoShapes", function(TachoColours, TachoBase, TachoShapes) {
  /**
   * Abstract
   * Represents the base app that all modded apps will use/extend
   */
  class BaseApp {
    /**
     * Constructor
     * {@param bngApi} is required as it cannot be requested here by AngularJS's injector for some
     * reason :/
     * @param {number} id - the ID of this app
     * @param {number} baseRadius - the radius of the outermost part of the app
     * @param {String} fontName - the name of the font to use
     * @param {[String]} units - the list of units that can be used for this app
     * @param {HTMLDocument} document - the document element the app uses
     * @param {?} bngApi - the bngApi service
     * @param {?} UiUnits - the UiUnits service
     */
    constructor(id, baseRadius, fontName, units, document, bngApi, UiUnits) {
      this.ID = id;
      this.FONT_NAME = fontName;
      this.UNITS = units;
      this.radius = baseRadius;
      this.CENTRE = new TachoBase.Vector(baseRadius, baseRadius);

      // Default to first
      this.unit = units[0];

      // A list of states the UI can be in
      this.states = Object.freeze({"preInit":1, "init":2, "loaded":3, "hidden":4});
      // The current state
      this.state = this.states.init;

      // The gauge object itself
      this.gauge = null;

      // The colour schemes available
      this.COLOUR_SCHEMES = TachoColours.COLOUR_SCHEMES;
      // The current colour scheme selected
      this.colourScheme = this.COLOUR_SCHEMES[0];

      // The foreground and background canvases
      this.CANVAS    = new TachoBase.Canvas(document, this.ID + "_canvas", this.radius*2);
      this.BG_CANVAS = new TachoBase.Canvas(document, this.ID + "_bgCanvas", this.radius*2);

      // The UiUnits service from BeamNG
      this.UiUnits = UiUnits;

      // Setup the background canvas's mouse action events
      this.initBGCanvas(bngApi);
    }

    /**
     * Sets up the background canvas, adding an event handler for the mouse action
     */
    initBGCanvas(bngApi) {
      let thisApp = this;

      this.CANVAS.setupCanvasClick(
          function(_event) {
            // Change unit
            thisApp.cycleUnit();
            thisApp.recreate(bngApi, true);
          },
          function(_event) {
            // Change colour scheme
            thisApp.cycleColourScheme();
            thisApp.recreate(bngApi, true);
          }
      );
    }

    /**
     * Recreates the gauge and re-draws the background canvas
     */
    recreate(bngApi, resave) {
      // Re-create gauge to reflect unit changes
      this.createGauge();
      // Show unit on background canvas
      this.drawBackgroundCanvas();

      if (resave)
          // Save this colour scheme and unit as default
        this.save(bngApi);
    }

    /**
     * Changes the current unit to the next available option
     */
    cycleUnit() {
      this.unit = BaseApp.cycleArray(this.unit, this.UNITS);
    }

    /**
     * Changes the current colour scheme to the next available option
     */
    cycleColourScheme() {
      this.colourScheme = BaseApp.cycleArray(this.colourScheme, this.COLOUR_SCHEMES);
    }

    /**
     * Saves the current colour scheme and unit to a file in the user's directory. Note that this
     * only saves the indexes with respect to the COLOUR_SCHEMES array and not the actual instance
     */
    save(bngApi) {
      let data = {
        unit: this.UNITS.indexOf(this.unit),
        colourScheme: this.COLOUR_SCHEMES.indexOf(this.colourScheme)
      }

      bngApi.engineLua("extensions.ui_tachometers.save('" + this.ID + "', " + bngApi.serializeToLua(data) + ")");
    }

    /**
     * Loads the settings file from the user's directory. If the file does not exist, this method
     * does nothing; else, it loads the unit and colour scheme to this app
     * @param {?} bngApi - the BeamNG API for interfacing with Lua
     */
    load(bngApi) {
      bngApi.engineLua("extensions.ui_tachometers.load('" + this.ID + "')", (response) => {
        if (!response)
          return;

        this.unit = this.UNITS[response.unit];
        this.colourScheme = this.COLOUR_SCHEMES[response.colourScheme];

        this.recreate(bngApi, false);
      });
    }

    /**
     * Draws to the main canvas
     */
    draw() {
      // this.prepCanvas(this.CANVAS);
      this.CANVAS.show(true);
      this.CANVAS.clear();
    }

    /**
     * Draws the app to the background canvas
     */
    drawBackgroundCanvas() {
      // this.prepCanvas(this.BG_CANVAS);
      this.BG_CANVAS.show(true);
      this.BG_CANVAS.clear();
    }

    /**
     * Hides the app by hiding both canvases
     */
    hide() {
      this.CANVAS.show(false);
      this.BG_CANVAS.show(false);
      this.state = this.states.hidden;
    }

    /**
     * Draws text in the colour scheme's primary text colour at the position specified
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     * @param {string} text - the text to write
     * @param {number} fontSize - the font size in px
     * @param {Vector} position - the position to set the centre of the text to
     * @param {boolean} [centerText] - whether to center the text (defaults to true)
     */
    drawPrimaryText(ctx, text, fontSize, position, centerText) {
      if (centerText === undefined)
        centerText = true;

      new TachoBase.Text(
          new TachoBase.Font(this.FONT_NAME, fontSize, centerText),
          text,
          position,
          this.colourScheme.TEXT.primary
      ).draw(ctx);
    }

    // prepCanvas(canvas) {
    //   let canvasElement = canvas.element;
    //   let dimensions = BaseApp.getObjectFitSize(
    //       canvasElement.clientWidth,
    //       canvasElement.clientHeight,
    //       canvasElement.width,
    //       canvasElement.height
    //   );
    //   canvasElement.width = dimensions.width;
    //   canvasElement.height = dimensions.height;
    //
    //   console.log(canvas, dimensions);
    //
    //   let ctx = canvas.get2DContext();
    //   let ratio = Math.min(
    //       canvasElement.clientWidth / this.radius,
    //       canvasElement.clientHeight / this.radius
    //   );
    //   ctx.scale(ratio, ratio);
    // }


    // adapted from: https://www.npmjs.com/package/intrinsic-scale
    static getObjectFitSize(containerWidth, containerHeight, width, height) {
      let doRatio = width / height;
      let cRatio = containerWidth / containerHeight;
      let test = doRatio > cRatio;

      let targetWidth;
      let targetHeight;

      if (test) {
        targetWidth = containerWidth;
        targetHeight = targetWidth / doRatio;
      } else {
        targetHeight = containerHeight;
        targetWidth = targetHeight * doRatio;
      }

      return {
        width: targetWidth,
        height: targetHeight
      };
    }

    /**
     * Returns the next value in an array, given a current value, looping around to the start if
     * the value is at the end of the array
     * @template T
     * @param {T} currValue - the current value
     * @param {[T]} array - the array containing a set of values, including {@param currValue}
     * @returns {T} the next value in {@param array}, succeeding {@param currValue}
     */
    static cycleArray(currValue, array) {
      // Find next index, looping back around to 0 if necessary
      let index = array.indexOf(currValue);
      index = (index+1) % array.length;

      return array[index];
    }

    /**
     * Resets the filter to prevent lingering filters
     */
    static resetFilter(ctx) {
      ctx.filter = "none";
      ctx.shadowBlur = 0;
      ctx.shadowColor = "none";
    }

    /**
     * Reloads the UI, hiding it after 50ms. If instantReload is true, this is done
     * instantly
     * @param {BoostApp} app - the app to reload
     * @param {boolean} instantReload - whether to instantly reload instead of waiting
     */
    static reload(app, instantReload) {
      console.log("Reloading");
      if (!app)
        return;

      let time = instantReload ? 0 : 50;
      setTimeout(() => {
        // It is unknown whether this vehicle will need this app, so hide it provisionally and
        // it is does, the app will re-show itself when updating occurs
        app.hide();
        app.state = app.states.init;
      }, time);
    }

    static changeSize(app, newWidth, newHeight) {
      // Chose smaller value to ensure fitting
      let radius = newWidth < newHeight ? newWidth : newHeight;

      app.CANVAS.element.width = radius;
      app.CANVAS.element.height = radius;

      app.BG_CANVAS.element.width = radius;
      app.BG_CANVAS.element.height = radius;

      console.log(radius, newWidth, newHeight);
    }


    /**
     * Creates the gauge element
     */
    createGauge() {
      console.error("createGauge() is abstract and must be implemented in the base class!");
    }
  }


  /**
   * Represents an vehicle's engine
   */
  class Engine {
    // The tolerance between when to shift and the maximum engine RPM
    static SHIFT_TOL = 200;

    /**
     * Constructor
     */
    constructor() {
      // The current RPM
      this.currRPM = 0;
      // The maximum engine RPM (as gotten from the vehicles stats)
      this.maxRPM = 0;
      // The next thousand from the maximum RPM, treated as the absolute maximum RPM
      this.redLineRPM = 0;
      // The oil temperature
      this.engineTemp = 0;
    }


    setCurrRPM(rpm) {
      if (typeof rpm !== "undefined")
        this.currRPM = rpm;
    }

    setMaxRPM(maxRPM) {
      if (typeof maxRPM !== "undefined")
        this.maxRPM = maxRPM;
    }

    setRedLineRPM(redLineRPM) {
      if (typeof redLineRPM !== "undefined")
        this.redLineRPM = redLineRPM;
    }

    setEngineTemp(engineTemp) {
      if (typeof engineTemp !== "undefined")
        this.engineTemp = engineTemp;
    }

    getCurrentRPM() {
      return this.currRPM;
    }

    getMaxRPM() {
      return this.maxRPM;
    }

    getRedLineRPM() {
      return this.redLineRPM;
    }

    getTemperature() {
      return this.engineTemp;
    }


    /**
     * @returns {number} the ratio of current RPM to redline RPM
     */
    getCurrentRPMRatio() {
      return this.currRPM / this.redLineRPM;
    }

    /**
     * @returns {number} the ratio of maximum engine RPM and the redline (next thousand) RPM
     */
    getRedlineRPMRatio() {
      return this.maxRPM / this.redLineRPM;
    }

    /**
     * @returns {boolean} whether to shift up to the next gear
     */
    shouldShiftUp() {
      return this.currRPM >= (this.maxRPM - Engine.SHIFT_TOL);
    }

    /**
     * @returns {number} the next thousand from the redline RPM
     */
    getNextRPMThousand() {
      return this.redLineRPM/1000 + 1
    }
  }

  /**
   * Represents a gearbox
   */
  class Gearbox {
    /**
     * Constructor
     * @param {string | number} gear - the currently selected gear
     */
    constructor(gear) {
      this.gear = gear;
    }

    /**
     * Sets the current gear
     * @param {string | number} gear - the gear to set
     */
    setGear(gear) {
      this.gear = gear;
    }

    /**
     * @returns {string | number} the gear number
     */
    getGear() {
      return this.gear;
    }

    /**
     * @return {string} Returns the string name of the gear currently selected on this gearbox
     */
    getGearName() {
      return typeof this.gear === "string" ? this.gear : this.getGearNameManual();
    }

    /**
     * Converts the gear number into its string name.
     * Mapping:
     *   < -1  -> R2...R(|n|+1)
     *   = -1  -> R
     *   =  0  -> N
     *  >=  1  -> 1...n
     * @returns {string} the string name for this gear number
     */
    getGearNameManual() {
      let gearName;

      if (this.gear === 0) {
        gearName = "N";
      } else if (this.gear < 0) {
        // Add index of reverse gear if more than one exists
        if (this.gear < -1)
          gearName = "R" + Math.abs(this.gear);
        else
          gearName = "R";
      } else {
        gearName = this.gear.toString();
      }

      return gearName;
    }
  }


  /**
   * Abstract
   * Represents a base Tachometer modded app
   */
  class TachometerApp extends BaseApp {
    /**
     * Constructor
     * @param {number} id - the ID of this app
     * @param {number} baseRadius - the radius of the outermost part of the app
     * @param {String} fontName - the name of the font to use
     * @param {Range} rpmAngleRange - the RPM angle range of the gauge
     * @param {HTMLDocument} document - the document element the app uses
     * @param {?} bngApi - the bngApi service
     * @param {?} UiUnits - the UiUnits service
     */
    constructor(id, rpmAngleRange, fontName, baseRadius, document, bngApi, UiUnits) {
      super(id, baseRadius, fontName, ["metric", "imperial"], document, bngApi, UiUnits);
      this.RPM_ANGLE_RANGE = rpmAngleRange;

      // The various attributes of the vehicle's state that may or may not be needed by the gauge
      this.lightsState = 0;
      this.speed = 0;
      this.indicators = {left: 0, right: 0};

      this.engine = new Engine();
      // @type Gearbox
      this.gearbox = null;

      this.pBrakeEngaged = false;
      this.fuel = 0;
    }

    /**
     * Updates the variables in the app's fields. If any variables have changed that affect the
     * background canvas, the flag needsRedraw is set to true.
     * @param {Object} engineInfo - the engineInfo stream
     * @param {Object} electrics - the electrics stream
     */
    updateAllVariables(engineInfo, electrics) {
      let fields = TachometerApp.getDynamicVars(engineInfo, electrics);

      this.setSpeed(fields.currSpeed);
      this.setLightsState(fields.lights);
      this.setParkingBrake(fields.pBrakeEngaged);
      this.setFuel(fields.fuel);

      this.setIndicatorsState(fields.indicators);

      this.engine.setCurrRPM(fields.engine.currRPM);
      this.engine.setEngineTemp(fields.engine.waterTemp);

      this.gearbox.setGear(fields.gearbox.currGear);
    }

    /**
     * Updates the variables in the app's fields. If any variables have changed that affect the
     * background canvas, the flag needsRedraw is set to true.
     * @param {Object} engineInfo - the engineInfo stream
     * @param {Object} electrics - the electrics stream
     */
    updateVariables(engineInfo, electrics) {
      let prevState = this.collateBGVars();
      this.updateAllVariables(engineInfo, electrics);
      let newState = this.collateBGVars();

      // If any variables for the background canvas have changed, the background canvas
      // requires a re-draw
      if (!prevState.every(function(v, i) {return v === newState[i]}))
        this.needsRedraw = true;
    }

    /**
     * Initialises the app's fields
     * @param {Object} engineInfo - the engineInfo stream
     * @param {Object} electrics - the electrics stream
     */
    initFields(engineInfo, electrics) {
      let gearbox = TachometerApp.getGearboxVars(engineInfo);
      this.gearbox = new Gearbox(gearbox.currGear);

      this.updateVariables(engineInfo, electrics);

      let maxRPM = TachometerApp.getEngineVars(engineInfo, electrics).maxRPM;
      this.engine.setMaxRPM(maxRPM);
      // Round to next thousand
      this.engine.setRedLineRPM(Math.floor(maxRPM/1000 + 1) * 1000);

      this.createGauge();
    }

    /**
     * @returns {[*]} the list of all variables that are used in the drawing on the background
     * canvas. This includes the gear index, the lights' state and the parking brake
     */
    collateBGVars() {
      return [this.lightsState, this.pBrakeEngaged, this.gearbox.getGear()];
    }

    setSpeed(speed) {
      if (typeof speed !== "undefined")
        this.speed = speed;
    }

    setLightsState(lightsState) {
      if (typeof lightsState !== "undefined" && lightsState >= 0 && lightsState <= 2)
        this.lightsState = lightsState;
    }

    setParkingBrake(pBrakeEngaged) {
      if (typeof pBrakeEngaged !== "undefined")
        this.pBrakeEngaged = pBrakeEngaged;
    }

    setFuel(fuel) {
      if (typeof fuel !== "undefined")
        this.fuel = fuel;
    }

    setIndicatorsState(indicators) {
      if (typeof indicators !== "undefined")
        this.indicators = indicators;
    }

    /**
     * @returns {number} the angle at which redline RPM occurs on the RPM gauge
     */
    getRedlineRPMAngle() {
      return this.RPM_ANGLE_RANGE.getLinearRatio(this.engine.getRedlineRPMRatio());
    }

    /**
     * @returns {number} the number of RPM thousands of the engine (including 0)
     */
    getNumRPMThousands() {
      // +1 to count 0
      return Math.floor(this.engine.getRedLineRPM()/1000) + 1
    }

    /**
     * Draws the RPM gauge. This includes the RPM ticks, pointer and current RPM arc
     */
    drawRPM() {
      let currRPMAngle = this.RPM_ANGLE_RANGE.getLinearRatio(this.engine.getCurrentRPMRatio());
      // Ensure the angle does not overflow past the boundaries
      currRPMAngle = this.RPM_ANGLE_RANGE.clampTo(currRPMAngle);

      this.gauge.setAngle(currRPMAngle);
      this.gauge.setValue(this.engine.getCurrentRPM());
      this.gauge.draw(this.CANVAS.get2DContext());
    }

    /**
     * Draws the indicators
     */
    drawIndicators() {
      this.drawIndicator(true, this.indicators.left > 0);
      this.drawIndicator(false, this.indicators.right > 0);
    }

    /**
     * Draws an indicator
     * @param {boolean} isLeft - whether this is the left indicator
     * @param {boolean} isOn - whether the indicator is currently active
     */
    drawIndicator(isLeft, isOn) {
      console.error("drawIndicator() MUST be implemented in base class!");
    }


    /**
     * Returns the relevant engine variables.
     * @param {Object} engineInfo - the engineInfo stream
     * @param {Object} electrics - the electrics stream
     * @returns {{maxRPM: number, currRPM: number, waterTemp: number}} the maximum and current RPM
     * of the engine and the water temperature inside
     */
    static getEngineVars(engineInfo, electrics) {
      return {
        currRPM: engineInfo[4],
        maxRPM: engineInfo[1],
        waterTemp: electrics.watertemp
      }
    }

    /**
     * Returns the relevant gearbox variables.
     * @param {Object} engineInfo - the engineInfo stream
     * @returns {{currGear: string | number}} the current gear
     */
    static getGearboxVars(engineInfo) {
      return {
        currGear: engineInfo[5]
      }
    }

    /**
     * Returns the state of the lights.
     * @param {Object} electrics - the electrics stream to read the light state
     * @returns {number} the state of the lights (0 = off, 1 = on, 2 = fog)
     */
    static getLightState(electrics) {
      if (electrics.lowbeam)
        return 1;
      else if (electrics.highbeam)
        return 2;
      else
        return 0;
    }

    /**
     * Returns the state of the indicators
     * @param {Object} electrics - the electrics stream to read the light state
     * @returns {{left: number, right: number}}
     */
    static getIndicators(electrics) {
      return {
        left: electrics.signal_L,
        right: electrics.signal_R,
      }
    }

    /**
     * Returns a list of all dynamic variables that will/may change between frames for the vehicle
     * @param {Object} engineInfo - the engine info object from the engineInfo stream
     * @param {Object} electrics - the electrics object from the electrics stream
     * @returns {Object} the list of variables
     */
    static getDynamicVars(engineInfo, electrics) {
      return {
        currSpeed: electrics.wheelspeed,
        lights: TachometerApp.getLightState(electrics),
        pBrakeEngaged: electrics.parkingbrake > 0,
        fuel: engineInfo[11] / engineInfo[12],
        engine: TachometerApp.getEngineVars(engineInfo, electrics),
        gearbox: TachometerApp.getGearboxVars(engineInfo),
        indicators: TachometerApp.getIndicators(electrics)
      };
    }

    /**
     * Sends and updates the vehicle's data streams to a tachometer app
     * @param {Object} streams - the vehicle's streams (should contain engineInfo and electrics
     * streams)
     * @param {TachometerApp} app - the app to update. Can be null, in which case, no update occurs.
     * If the app is also not ready or the streams are not ready, no update occurs.
     * @param {Object} bngApi - the bngApi
     */
    static sendStreamEvent(streams, app, bngApi) {
      let engineInfo = streams.engineInfo;
      let electrics = streams.electrics;

      // Ensure engineInfo exists AND ALSO that the engine is actually setup at this point
      let isEngineSetup = engineInfo && TachometerApp.getEngineVars(engineInfo, electrics).maxRPM !== 0;

      // Wait until everything is ready
      if (!app || app.state === app.states.preInit || !isEngineSetup || !electrics)
        return;

      // Perform first time setup if necessary
      if (app.state === app.states.init) {
        app.initFields(engineInfo, electrics);
        app.load(bngApi);
        app.drawBackgroundCanvas();
        app.state = app.states.loaded;
      }

      app.updateVariables(engineInfo, electrics);
      app.draw();
    }
  }

  /**
   * Abstract
   * Represents a base Boost modded app
   */
  class BoostApp extends BaseApp {
    /**
     * Constructor
     * @param {number} id - the ID of this app
     * @param {number} baseRadius - the radius of the outermost part of the app
     * @param {String} fontName - the name of the font to use
     * @param {Range} boostAngleRange - the boost angle range of the gauge
     * @param {HTMLDocument} document - the document element the app uses
     * @param {?} bngApi - the bngApi service
     * @param {?} UiUnits - the UiUnits service
     */
    constructor(id, boostAngleRange, fontName, baseRadius, document, bngApi, UiUnits) {
      super(id, baseRadius, fontName, ["inHg", "bar", "psi", "kPa"], document, bngApi, UiUnits);
      this.BOOST_ANGLE_RANGE = boostAngleRange;

      this.maxBoost = 0;
      this.currBoost = 0;
    }

    /**
     * Initialises the app's fields.
     * Note that both values should be their 'raw' values, so that they can be converted to specific
     * units later
     * @param {number} currBoost - the current boost of the turbo
     * @param {number} maxBoost - the maximum possible boost value
     */
    initFields(currBoost, maxBoost) {
      this.setMaxBoost(maxBoost);
      this.setCurrBoost(currBoost);

      this.createGauge();
    }

    setMaxBoost(maxBoost) {
      if (typeof maxBoost !== "undefined")
        this.maxBoost = maxBoost;
    }

    setCurrBoost(currBoost) {
      if (typeof currBoost !== "undefined")
        this.currBoost = currBoost;
    }

    /**
     * @returns {number} the current boost value in the current unit
     */
    getUnitisedBoostValue() {
      return this.UiUnits.pressure(this.currBoost, this.unit).val;
    }
  }

  this.BaseApp = BaseApp;
  this.TachometerApp = TachometerApp;
  this.BoostApp = BoostApp;

  return this;
}])