angular.module("beamng.apps")
/**
 * The base service, providing miscellaneous classes and functions that all Tachometer apps and
 * other base services are likely to use and require
 */
.factory("TachoBase", [function() {
  /**
   * A class representing a 2D vector
   */
  class Vector {
    constructor(x, y) {
      this.array = [x || 0, y || 0];
    }

    /**
     * Copies a vector and returns its copy
     * @param {Vector} v - the vector to copy
     * @returns {Vector} the copied vector instance
     */
    static copy(v) {
      return new Vector(v.array[0], v.array[1]);
    }

    /**
     * Reverse the signs of all elements in the vector
     */
    negate() {
      this.array.forEach(e => -e);
    }

    /**
     *
     * @param {Vector} vector
     * @param {Vector|number} amount
     * @param {function} func
     * @returns {Vector}
     */
    static mapOperations(vector, amount, func) {
      let result = new Vector(vector.array[0], vector.array[1]);

      if (amount instanceof Vector) {
        for (let i = 0; i < result.array.length; i++)
          result.array[i] = func(result.array[i], amount.array[i]);
      } else {
        result.array = result.array.map((e) => func(e, amount));
      }

      return result;
    }

    /**
     * Adds a vector from another or a scalar value
     * @param {Vector} vector - the vector to be added to
     * @param {Vector|number} amount - the vector/scalar to add
     * @returns {Vector} - the resulting vector
     */
    static add(vector, amount) {
      return Vector.mapOperations(vector, amount, (e, a) => e + a);
    }

    /**
     * Subtracts a vector from another or a scalar value
     * @param {Vector} vector - the vector to be subtracted from
     * @param {Vector|number} amount - the vector/scalar to subtract
     * @returns {Vector} - the resulting vector
     */
    static sub(vector, amount) {
      return Vector.mapOperations(vector, amount, (e, a) => e - a);
    }

    /**
     * Multiples a vector with another or a scalar value
     * @param {Vector} vector - the vector to be subtracted from
     * @param {Vector|number} amount - the vector/scalar to subtract
     * @returns {Vector} - the resulting vector
     */
    static mult(vector, amount) {
      return Vector.mapOperations(vector, amount, (e, a) => e * a);
    }

    /**
     * Divides a vector by another or a scalar value
     * @param {Vector} vector - the vector to be divided by
     * @param {Vector|number} amount - the vector/scalar to divide with
     * @returns {Vector} - the resulting vector
     */
    static div(vector, amount) {
      return Vector.mapOperations(vector, amount, (e, a) => e / a);
    }

    /**
     * Adds to this vector another vector from another or a scalar value
     * @param {Vector|number} amount - the vector/scalar to add
     */
    add(amount) {
      this.array = Vector.add(this, amount).array;
    }

    /**
     * Subtracts this vector by another or a scalar value
     * @param {Vector|number} amount - the vector/scalar to subtract
     */
    sub(amount) {
      this.array = Vector.sub(this, amount).array;
    }

    /**
     * Multiples this vector with another or a scalar value
     * @param {Vector|number} amount - the vector/scalar to subtract
     */
    mult(amount) {
      this.array = Vector.mult(this, amount).array;
    }

    /**
     * Divides this vector by another or a scalar value
     * @param {Vector|number} amount - the vector/scalar to divide with
     */
    div(amount) {
      this.array = Vector.div(this, amount).array;
    }

    /**
     * Normalises this vector
     */
    normalise() {
      let mag = this.mag();

      if (mag > 0)
        this.div(mag);
    }

    /**
     * Calculates and returns this vector's magnitude
     * @returns {number} the magnitude of this vector
     */
    mag() {
      return Math.sqrt(this.array.reduce((a, b) => Math.pow(a, 2) + Math.pow(b, 2)));
    }

    /**
     * Sets this vector's components to that of another's
     * @param {Vector} vector - the components to copy
     */
    set(vector) {
      for (let i = 0; i < this.array.length; i++)
        this.array[i] = vector.array[i];
    }

    setX(x) {
      this.array[0] = x;
    }

    setY(y) {
      this.array[1] = y;
    }

    addX(incX) {
      this.array[0] += incX;
    }

    addY(incY) {
      this.array[1] += incY;
    }

    x() {
      return this.array[0];
    }

    y() {
      return this.array[1];
    }
  }

  /**
   * Represents a canvas object that can be drawn on
   */
  class Canvas {
    /**
     * Constructor
     * @param {HTMLDocument} document - the document object this canvas is on
     * @param {string} id - the canvas's identifier string
     * @param {number} dims - the width and height of the canvas
     */
    constructor(document, id, dims) {
      this.element = document.getElementById(id);
      this.ID = id;

      this.setupCanvas(dims);
    }

    /**
     * Ensures proper HD scaling of text to eliminate blurry text and geometry
     */
    setupCanvas(dims) {
      // Ensure correct HD resolution on the canvas (fix for blurry text)
      this.element.width = dims;
      this.element.height = dims;
      this.get2DContext().imageSmoothingEnabled = true;
      this.show(false);
    }

    /**
     * Shows or hides the canvas
     * @param {boolean} show - whether to show or hide the canvas
     */
    show(show) {
      this.element.style.opacity = show ? "1" : "0";
    }

    /**
     * @returns {CanvasRenderingContext2D} the 2D context of the canvas
     */
    get2DContext() {
      return this.element.getContext("2d");
    }

    /**
     * Clears the canvas of all content drawn on it
     */
    clear() {
      let ctx = this.get2DContext();
      ctx.clearRect(0, 0, this.element.width, this.element.height);
    }

    /**
     * Sets a listener to a canvas for the left and right click action by a mouse
     * @param {function} leftFunc - the function to perform on a left click
     * @param {function} rightFunc - the function to perform on a right click
     */
    setupCanvasClick(leftFunc, rightFunc) {
      this.element.addEventListener("mousedown", function(event) {
        switch (event.button) {
          case 0:
            leftFunc(event);
            break;
          case 2:
            rightFunc(event);
            break;
        }
      });
    }

    /**
     * Resets the transformation of a canvas's 2D context
     * @param {CanvasRenderingContext2D} ctx - the context to reset the transformation of
     */
    static resetTransformationState(ctx) {
      ctx.setTransform(1, 0, 0, 1, 0, 0);
    }
  }

  /**
   * Represents a range of values between two values
   */
  class Range {
    /**
     * Constructor
     * The lower bound does not have to be smaller than the upper bound, and in which case, the
     * range is considered reversed
     * @param {number} lower - the lower bound of the range
     * @param {number} upper - the upper bound of the range
     */
    constructor(lower, upper) {
      this.lower = lower;
      this.upper = upper;
    }

    /**
     * Inverts the bounds such that the upper and lower bound swap
     * @returns {Range}
     */
    invert() {
      return new Range(this.upper, this.lower);
    }

    setLower(lower) { this.lower = lower; }
    setUpper(upper) { this.upper = upper; }

    /**
     * Pads the bounds outwards by a set amount. This lowers the lower bound by the set amount and
     * increases the upper bound by the set amount
     * @param {number} amount - the amount to pad
     */
    pad(amount) {
      this.lower -= amount;
      this.upper += amount;
    }

    /**
     * Clamps a value between this number range, such that a value lesser than the lower bound
     * will be set to the lower bound and vice versa for the upper bound. A value which lies
     * within the bounds is not affected
     * @param {number} value - the value to clamp
     * @returns {number} the clamped value
     */
    clampTo(value) {
      return Math.min(Math.max(value, this.lower), this.upper);
    }

    /**
     * Returns the value of the specified distance along this number range.
     * Example: {lower=5, upper=10}.getLinearRatio(0.5) -> 7.5
     * @param distance - the distance along the range
     * @returns {number} the value at {distance} along this range
     */
    getLinearRatio(distance) {
      return this.lower + distance * (this.upper - this.lower);
    }

    /**
     * @returns {number} the differences between the two bounds
     */
    getDifference() {
      return this.upper - this.lower;
    }

    clone() {
      return new Range(this.lower, this.upper);
    }
  }

  class LinearGradient {
    /**
     * Constructor
     * @param {Vector} startPos - the start position of the gradient
     * @param {Vector} endPos - the end position of the gradient
     * @param {[string]} colours - the colours in the gradient
     */
    constructor(startPos, endPos, colours) {
      this.startPos = startPos;
      this.endPos = endPos;
      this.colours = colours;
    }

    apply(ctx) {
      let grad = ctx.createLinearGradient(
          this.startPos.x(), this.startPos.y(), this.endPos.x(), this.endPos.y()
      );

      for (let i = 0; i < this.colours.length; i++)
        grad.addColorStop(i / (this.colours.length-1), this.colours[i]);

      ctx.fillStyle = grad;
    }
  }


  /**
   * Represents a base style
   * Abstract/Interface
   */
  class BaseStyle {
    constructor() {}

    /**
     * Applies styling to a drawn-on canvas context and renders it
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    apply(ctx) {
      console.error("apply() MUST be implemented in base class!")
    }
  }

  /**
   * Represents a style which can be applied to shapes and objects being drawn on a canvas that
   * has a gradient as its fill
   */
  class GradientStyle extends BaseStyle {
    /**
     * Constructor
     * @param {LinearGradient} [gradient] - a gradient to apply
     */
    constructor(gradient) {
      super();
      this.gradient = gradient;
    }

    /**
     * @Override
     * @inheritDoc
     */
    apply(ctx) {
      this.gradient.apply(ctx);
      ctx.fill();
    }
  }

  /**
   * Represents a style which can be applied to shapes and objects being drawn on a canvas
   */
  class Style extends BaseStyle {
    /**
     * Constructor
     * @param {string} strokeColour - the colour of the stroke. Can be either RGB/RGBA or colour
     * name
     * @param {number} lineWidth - the width of the line/outline
     * @param {boolean} [filled] - whether the shape will be filled (fill colour will be the same as stroke
     * colour). Defaults to false
     */
    constructor(strokeColour, lineWidth, filled) {
      super();
      this.strokeColour = strokeColour;
      this.lineWidth = lineWidth;
      this.filled = filled || false;
    }

    /**
     * Returns a Style object of the specified colour that has a 1px stroke and is filled
     * @param {string} strokeColour - the colour of the stroke. Can be either RGB/RGBA or colour
     * name
     * @returns {Style} the style object created
     */
    static unitStrokeFill(strokeColour) {
      return new Style(strokeColour, 1, true);
    }

    /**
     * Returns a Style object that is filled with the specified colour and that does not have a
     * stroke applied to it
     * @param {string} fillColour - the fill colour. Can be either RGB/RGBA or colour name
     * @returns {Style} the style object created
     */
    static noStrokeFill(fillColour) {
      return new Style(fillColour, 0, true);
    }

    /**
     * @Override
     * @inheritDoc
     */
    apply(ctx) {
      if (this.filled) {
        ctx.fillStyle = this.strokeColour;
        ctx.fill();
        return;
      }

      ctx.strokeStyle = this.strokeColour;
      ctx.lineWidth = this.lineWidth;
      ctx.stroke();
    }
  }


  /**
   * Abstract
   * Represents an object is drawable to a canvas context object
   */
  class Drawable {
    constructor() {}

    /**
     * Draws the object to a canvas context object
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    draw(ctx) {
      console.error("draw() must be implemented by class!");
    }
  }

  /**
   * Abstract
   * Represents an object that can be moved
   */
  class Moveable extends Drawable {
    /**
     * Constructor
     * @param {Vector} [centre] - the centre point of the object. Defaults to (0,0)
     */
    constructor(centre) {
      super();
      this.centre = centre || new Vector();
    }

    /**
     * @param {Vector} centre - the new centre to set the object to
     */
    setCentre(centre) { this.centre = centre; };
  }

  /**
   * Abstract
   * Represents an object that can be rotated around a centre point
   */
  class Rotatable extends Moveable {
    /**
     * Constructor
     * @param {Vector} [centre] - the rotational centre. Defaults to (0,0)
     * @param {number} [angle] - the angle around the centre. Defaults to 0
     */
    constructor(centre, angle) {
      super(centre);
      this.angle = angle || 0;
    }

    /**
     * @param {number} angle - the angle to set
     */
    setAngle(angle) { this.angle = angle; };
  }



  /**
   * Represents a font for text displaying
   */
  class Font {
    /**
     * Constructor
     * @param {string} fontName - the font name
     * @param {number} fontSize - the font size
     * @param {boolean} isCentred - whether the text is centre
     */
    constructor(fontName, fontSize, isCentred) {
      this.fontName = fontName;
      this.fontSize = fontSize;
      this.isCentred = isCentred;
    }

    apply(ctx) {
      ctx.textAlign = this.isCentred ? "center" : "left";
      ctx.textBaseline = this.isCentred ? "middle" : "alphabetic";
      ctx.font = this.fontSize + "px " + this.fontName;
    }
  }

  /**
   * Represents text of a given font and position
   */
  class Text extends Drawable {
    /**
     * Constructor
     * @param {Font} font - the font to use for the text
     * @param {string} text - the text to display
     * @param {Vector} position - the position the text will be displayed at
     * @param {string} [colour] - the text colour (optional, leave omit for transparent text (
     * gradients will fill in here))
     */
    constructor(font, text, position, colour) {
      super();
      this.font = font;
      this.text = text;
      this.position = position;
      this.colour = colour;
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      this.font.apply(ctx);

      if (this.colour)
        ctx.fillStyle = this.colour;

      ctx.fillText(this.text, this.position.x(), this.position.y());
    }
  }



  this.Vector = Vector;

  this.Canvas = Canvas;

  this.Drawable = Drawable;
  this.Moveable = Moveable;
  this.Rotatable = Rotatable;

  this.Font = Font;
  this.Text = Text;
  this.Range = Range;
  this.LinearGradient = LinearGradient;
  this.GradientStyle = GradientStyle;
  this.Style = Style;


  return this;
}])