angular.module("beamng.apps")
/**
 * Provides colour classes and some additional colour related functions
 */
.factory("TachoColours", [function() {
  /**
   * Represents a dual-colour scheme with a primary/default colour and a secondary/backup colour
   */
  class ColourPriority {
    /**
     * Constructor
     * @param {string} primary - the primary colour
     * @param {string} secondary - the secondary colour
     */
    constructor(primary, secondary) {
      this.primary = primary;
      this.secondary = secondary;
    }
  }

  /**
   * Represents a colour scheme for the UI
   */
  class ColourScheme {
    /**
     * Constructor
     * @param {string} base - the base colour, used for the main body
     * @param {string} background - the background elements' colour
     * @param {ColourPriority} text - the text colour
     * @param {ColourPriority} accent - the accent colour
     */
    constructor(base, background, text, accent) {
      this.BASE = base;
      this.BACKGROUND = background;
      // @type ColourPriority
      this.TEXT = text;
      // @type ColourPriority
      this.ACCENT = accent;
    }
  }

  /**
   * Represents an RGB colour
   */
  class ColourRGB {
    constructor(r, g, b) {
      if (typeof r === "object" && !g && !b)
        this.rgb = r.map((x) => x);
      else
        this.rgb = [r, g, b];
    }

    toString() {
      return this.rgb[0] + ", " + this.rgb[1] + ", " + this.rgb[2];
    }
  }



  function createColourScheme(rgb1, rgb2, rgb3) {
    // This is constant for the colour schemes
    const GREY_SCALE_SCHEME = {
      BASE: "rgb(40, 40, 40)",
      SECONDARY_TEXT: "rgb(255, 255, 255)",
      ACCENT: "rgb(0, 0, 0)"
    }

    rgb2 = rgb2 || rgb1;
    rgb3 = rgb3 || rgb1;

    return new ColourScheme(
        GREY_SCALE_SCHEME.BASE,
        "rgba(" + rgb1.toString() + ", 0.75)",
        new ColourPriority("rgb(" + rgb2.toString() + ")", GREY_SCALE_SCHEME.SECONDARY_TEXT),
        new ColourPriority(GREY_SCALE_SCHEME.ACCENT, "rgb(" + rgb3.toString() + ")"),
    )
  }

  function generateColourSchemes() {
    let colours = {
      red      : new ColourRGB(255,   0,   0),
      orange   : new ColourRGB(255, 127,   0),
      green    : new ColourRGB(  0, 255,   0),
      blue     : new ColourRGB(  0, 100, 255),
      yellow   : new ColourRGB(255, 255,   0),
      lightblue: new ColourRGB(  0, 255, 255),
      purple   : new ColourRGB(150,   0, 255),
      grey     : new ColourRGB(100, 100, 100),
      white    : new ColourRGB(255, 255, 255)
    }

    let colourSchemes = [];

    // Generate tri-coloured colour schemes
    // BeamNG
    colourSchemes.push(createColourScheme(colours.grey, colours.orange, colours.orange));
    // Fire
    colourSchemes.push(createColourScheme(colours.red, colours.yellow, colours.orange));
    // Arsenic
    colourSchemes.push(createColourScheme(colours.green, colours.yellow, colours.green));
    // Crystal
    colourSchemes.push(createColourScheme(colours.blue, colours.lightblue, colours.purple));

    // Generate uni-coloured colour schemes
    for (let key in colours) {
      // Ignore grey as too dark
      if (key === "grey")
        continue;

      colourSchemes.push(createColourScheme(colours[key]));
    }


    return colourSchemes;
  }


  this.ColourPriority = ColourPriority;
  this.ColourScheme = ColourScheme;
  this.COLOUR_SCHEMES = generateColourSchemes();

  return this;
}])