angular.module("beamng.apps")
/**
 * Provides gauge classes for constructing the tachometer and boost gauges
 */
.factory("TachoGauge", ["TachoBase", "TachoShapes", function(TachoBase, TachoShapes) {

  /**
   * Abstract
   * Represents a gauge element which displays a set of objects onto the gauge
   */
  class GaugeArrayElement extends TachoBase.Moveable {
    /**
     * Constructor
     * @param {number} numElements - the number of elements
     * @param {number} valueSpacing - the value spacing between adjacent numbers
     * @param {Range} angleRange - the starting and ending angle range of the elements
     */
    constructor(numElements, valueSpacing, angleRange) {
      super();
      this.numElements = numElements;
      this.valueSpacing = valueSpacing;
      this.angleRange = angleRange;
      // -1 as element 0 starts at angleRange.lower and element N ends at angleRange.upper
      // so include upper bound as element N (not N+1)
      this.angleSpacing = angleRange.getDifference() / (numElements-1);
      this.elements = [];
    }

    /**
     * Abstract
     * Generates the elements
     * @return {?} the elements
     */
    generateElements() {
      console.error("generateElements() must be implemented in subclass!")
    }
  }

  /*
   * Represents the gauge's ticks that line the gauge
   */
  class GaugeTicks extends GaugeArrayElement {
    /**
     * Constructor
     * @param {number} numTicks - the number of ticks on gauge
     * @param {number} spacing - the spacing between two ticks, in terms of value
     * @param {number} width - the stroke width of each tick
     * @param {number} majorSpacing - the number of ticks between two major ticks
     * @param {ColourPriority} colour - the colour to apply to the ticks
     * @param {Range} tickRadii - the inner and outer radii of the ticks
     * @param {Range} angles - the angle of the first tick and the angle of the last tick
     * @param {boolean} [drawEndTicks] - whether to draw the ticks at either of the ends. Defaults
     * to false
     * @param {number} [minTickVal] - the lowest tick's value. Defaults to 0
     */
    constructor(numTicks, spacing, width, majorSpacing, colour, tickRadii, angles, drawEndTicks, minTickVal) {
      super(numTicks, spacing, angles);
      this.majorSpacing = majorSpacing;
      this.width = width;
      this.colour = colour;
      this.tickRadii = tickRadii;
      this.drawEndTicks = drawEndTicks || false;
      this.minTickVal = minTickVal || 0;

      this.elements = this.generateElements();
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      let oldCentre = this.centre;
      super.setCentre(centre);
      let translation = TachoBase.Vector.sub(this.centre, oldCentre);

      // Translate the ticks to the new centre point
      for (let tick of this.elements) {
        tick[0].add(translation);
        tick[1].add(translation);
      }
    }

    /**
     * @Override
     * @inheritDoc
     * Generates a list of ticks as pairs of radii
     * @returns {[[Vector, Vector]]} the ticks, as pairs of inner and outer radii
     */
    generateElements() {
      let ticks = [];
      let currAngle = this.angleRange.lower;

      for (let i = 0; i < this.numElements; i++) {
        let isMajor = i % this.majorSpacing !== 0;
        let isEndTick = i === 0 || i === this.numElements-1;

        // Only draw end ticks if wanted
        if (!isEndTick || this.drawEndTicks)
          ticks.push(this.addTick(currAngle, isMajor));

        currAngle += this.angleSpacing;
      }

      return ticks;
    }

    /**
     * Returns an RPM tick, as a pair of radii, of the given inputs. The radii returned are the
     * outer and inner radii of the tick.
     * @param {number} angle - the angle this tick sits at around the RPM
     * @param {boolean} isMinor - whether this tick is representing a minor point on the gauge (ie.
     * a thousand on an RPM gauge)
     * @returns {[Vector, Vector]} the radii of the tick
     */
    addTick(angle, isMinor) {
      let r1 = this.getTickRadius(false);
      let r2 = this.getTickRadius(true);
      // Half size for minor ticks
      if (isMinor)
        r2 += (r1-r2)/2;

      return [
        TachoShapes.getRelCircularPoint(this.centre, r1, angle),
        TachoShapes.getRelCircularPoint(this.centre, r2, angle)
      ];
    }

    /**
     * Returns the radius on the RPM gauge a tick will start/end at
     * @param {boolean} isInner - whether to return the inner or outer radius
     * @returns {number}
     */
    getTickRadius(isInner) {
      return isInner ? this.tickRadii.lower : this.tickRadii.upper;
    }

    /**
     * Returns the value at the specified tick index
     * @param {number} index - the tick index
     * @returns {number} the value at that tick index
     */
    getValueAtIndex(index) {
      // If the start tick is not being drawn, index 0 is actually index 1
      let actualIndex = this.drawEndTicks ? index : index + 1;
      return this.minTickVal + actualIndex * this.valueSpacing;
    }

    /**
     * Draws ticks on the gauge as a disconnected series of lines. If the current value is greater
     * than the tick, the secondary accent colour will be applied; else, the primary accent will
     * be applied
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     * @param {number} currValue - the current value the angle on the gauge corresponds to
     */
    drawTicks(ctx, currValue) {
      for (let i = 0; i < this.elements.length; i++) {
        let tick = this.elements[i];
        new TachoShapes.Line(tick[0], tick[1], new TachoBase.Style(this.colour.secondary, this.width))
            .draw(ctx);
      }
    };
  }

  /**
   * Represents the gauge's ticks that line the gauge. If tick is below a particular value, that
   * tick is coloured differently, using the colour schemes primary colour.
   */
  class BiColourGaugeTicks extends GaugeTicks {
    /**
     * Constructor
     * @param {number} numTicks - the number of ticks on gauge
     * @param {number} spacing - the spacing between two ticks, in terms of value
     * @param {number} width - the stroke width of each tick
     * @param {number} majorSpacing - the number of ticks between two major ticks
     * @param {ColourPriority} colour - the colour to apply to the ticks
     * @param {Range} tickRadii - the inner and outer radii of the ticks
     * @param {Range} angles - the angle of the first tick and the angle of the last tick
     * @param {boolean} [drawEndTicks] - whether to draw the ticks at either of the ends. Defaults
     * to false
     * @param {number} [minTickVal] - the lowest tick's value. Defaults to 0
     */
    constructor(numTicks, spacing, width, majorSpacing, colour, tickRadii, angles, drawEndTicks, minTickVal) {
      super(numTicks, spacing, width, majorSpacing, colour, tickRadii, angles, drawEndTicks, minTickVal);
    }


    /**
     * Returns the accent colour of a tick based on whether it has been covered by the current
     * value on the gauge
     * @param {boolean} isValPastTick - whether the tick being drawn is covered by the current
     * value on the gauge
     * @returns {string} the accent colour
     */
    getAccentColour(isValPastTick) {
      return isValPastTick ? this.colour.primary : this.colour.secondary;
    }

    /**
     * @Override
     * @inheritDoc
     * If the current value is greater than the tick, the secondary accent colour will be applied;
     * else, the primary accent will be applied
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     * @param {number} valThresh - value threshold at which ticks smaller than or equal to are
     * coloured using the primary colour and vice versa for the secondary colour.
     */
    drawTicks(ctx, valThresh) {
      for (let i = 0; i < this.elements.length; i++) {
        let tick = this.elements[i];
        // If current value is past this tick, make it a different colour
        let colour = this.getAccentColour(valThresh >= this.getValueAtIndex(i));

        new TachoShapes.Line(tick[0], tick[1], new TachoBase.Style(colour, this.width))
            .draw(ctx);
      }
    };
  }

  /**
   * Represents a gauges numbering system that lines the gauge and displays that gauge's value at
   * set intervals
   */
  class GaugeNumbers extends GaugeArrayElement {
    /**
     * Constructor
     * @param {number} numNumbers - the number of gauge numbers
     * @param {number} spacing - the numerical spacing between adjacent gauge numbers
     * @param {Range} angles - the start and end angles the gauge numbers will start and end at
     * @param {Font} font - the font to use for displaying the gauge numbers
     * @param {number} radius - the radius of the circle the gauge numbers will wrap around from
     * @param {string} textColour - the colour of the gauge numbers
     * @param {number} [startNum] - the number to start at. Defaults to 0
     */
    constructor(numNumbers, spacing, angles, font, radius, textColour, startNum) {
      super(numNumbers, spacing, angles);
      this.font = font;
      this.radius = radius;
      this.textColour = textColour;
      this.startNum = startNum || 0;

      this.elements = this.generateElements();
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      let oldCentre = this.centre;
      super.setCentre(centre);
      let translation = TachoBase.Vector.sub(this.centre, oldCentre);

      // Translate the ticks to the new centre point
      for (let number of this.elements)
        number.position.add(translation);
    }

    /**
     * @Override
     * @inheritDoc
     * @returns {[Text]} the numbers generated
     */
    generateElements() {
      let numbers = [];
      let angle = this.angleRange.lower;

      for (let i = 0; i < this.numElements; i++) {
        let value = this.startNum + (i * this.valueSpacing);
        numbers.push(this.generateNumber(Math.trunc(value), angle));

        angle += this.angleSpacing;
      }

      return numbers;
    }

    /**
     * Creates a number in the form of a Text object
     * @param {number} number - the
     * @param angle
     * @returns {Text}
     */
    generateNumber(number, angle) {
      let position = TachoShapes.getRelCircularPoint(this.centre, this.radius, angle);
      return new TachoBase.Text(this.font, number.toString(), position, this.textColour);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      this.elements.forEach((e) => e.draw(ctx));
    }
  }


  class GaugeValueElement extends TachoBase.Moveable {
    /**
     * Constructor
     */
    constructor() {
      super();
    }

    /**
     * Sets the angle of the active part of the element
     * @param {number} angle - the angle to set
     */
    setActiveAngle(angle) {
      console.error("setActiveAngle() must be implemented in base class!");
    }
  }

  /**
   * Represents the arc(s) that wrap around the outside of a gauge
   */
  class GaugeArcs extends GaugeValueElement {
    /**
     * Constructor
     * @param {Arc} [mainArc] - the main arc which will represent the current value
     * @param {Arc} [endArc] - the arc that sits at the end. This is optional, and an example would
     * be the redline buffer for an engine or boost gauge
     * @param {Arc} [startArc] - the arc that sits at the start. This is optional
     */
    constructor(mainArc, endArc, startArc) {
      super();
      this.mainArc = mainArc;
      this.endArc = endArc;
      this.startArc = startArc;
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      if (this.endArc)
        this.endArc.setCentre(centre);
      if (this.startArc)
        this.startArc.setCentre(centre);
      if (this.mainArc)
        this.mainArc.setCentre(centre);
    }

    /**
     * @Override
     * @inheritDoc
     */
    setActiveAngle(angle) {
      this.mainArc.setEndAngle(angle);
    }

    /**
     * @Override
     * @inheritDoc
     * It should be noted that the start and end arcs (if they exist) have higher drawing priority
     * over the main arc and so will always be shown on top
     */
    draw(ctx) {
      if (this.mainArc)
        this.mainArc.draw(ctx);

      if (this.endArc)
        this.endArc.draw(ctx);
      if (this.startArc)
        this.startArc.draw(ctx);
    }
  }


  /**
   * Abstract
   */
  class GaugePointer extends GaugeValueElement {
    constructor() {
      super();
    }

    /**
     * Returns the angle 180deg added to the current angle. Note that the value returned will
     * always be +180deg from the current, regardless of whether this makes the angle bigger than
     * 360deg
     * @param {number} angle - the angle to get the inverse of
     * @returns {number} the angle 180deg plus the current angle, in radians
     */
    getInvAngle(angle) {
      return angle + Math.PI;
    }

    /**
     * Sets the centre of rotation for the pointers
     * @param {Vector} centre - the centre of rotation to set
     */
    setRotationalCentre(centre) {
      console.log("setRotationalCentre() must be implemeneted in base class!");
    }
  }

  /**
   * Represents a gauge pointer as a collection of the centre dial, minor pointer and major pointer
   */
  class SimplePointer extends GaugePointer {
    /**
     * Constructor
     * @param {Shape} majorPointer - the major part of the pointer
     * @param {Shape} minorPointer - the minor part of the pointer (points to the inverse angle)
     * @param {Shape} pointerDial - the pointer dial at the centre
     */
    constructor(majorPointer, minorPointer, pointerDial) {
      super();
      this.majorPointer = majorPointer;
      this.minorPointer = minorPointer;
      this.pointerDial = pointerDial;
    }

    /**
     * @Override
     * @inheritDoc
     */
    setActiveAngle(angle) {
      // An angle of 0 means the pointer is at the bottom -> -Math.PI/2
      angle -= Math.PI/2;

      this.majorPointer.setAngle(angle);
      this.minorPointer.setAngle(this.getInvAngle(angle));
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      super.setCentre(centre);
      this.majorPointer.setCentre(centre);
      this.minorPointer.setCentre(centre);
    }

    /**
     * @Override
     * @inheritDoc
     */
    setRotationalCentre(centre) {
      this.majorPointer.setRotationalCentre(centre);
      this.minorPointer.setRotationalCentre(centre);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      this.drawPointers(ctx);
      this.pointerDial.draw(ctx);
    }

    /**
     * Draws the pointers
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    drawPointers(ctx) {
      this.majorPointer.draw(ctx);
      this.minorPointer.draw(ctx);
    }
  }

  /**
   * Represents a pointer object, wherein an accent is displayed over the pointer
   */
  class AccentPointer extends GaugePointer {
    /**
     * Constructor
     * The angle this pointer points at is initially set to 0
     * @param {SimplePointer} bgPointer - the background pointer
     * @param {SimplePointer} accentPointer - the accent pointer
     * @param {Circle} pointerDial - the centre dial part of the pointers
     */
    constructor(bgPointer, accentPointer, pointerDial) {
      super();
      this.bgPointer = bgPointer;
      this.accentPointer = accentPointer;
      this.pointerDial = pointerDial;
    }

    /**
     * @Override
     * @inheritDoc
     */
    setActiveAngle(angle) {
      this.accentPointer.setActiveAngle(angle);
      this.bgPointer.setActiveAngle(angle);
    }

    /**
     * @Override
     * @inheritDoc
     */
    setRotationalCentre(centre) {
      this.accentPointer.setRotationalCentre(centre);
      this.bgPointer.setRotationalCentre(centre);
    }

    /**
     * @Override
     * Draws the pointer pieces to a canvas.
     * Note that pointerDial overrides the pointer dials specified in the pointer objects
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    draw(ctx) {
      this.bgPointer.drawPointers(ctx);
      this.accentPointer.drawPointers(ctx);
      this.pointerDial.draw(ctx);
    }
  }



  /**
   * Represents a gauge. This gauge has a numbering system and also a tick system to count notable
   * values along the gauge. One (or optionally two) GaugeValueElements show the current value,
   * either through a classical pointer, arc system or otherwise.
   */
  class Gauge extends TachoBase.Drawable {
    /**
     * Constructor
     * @param {GaugeNumbers} numbering - the numbers along the gauge
     * @param {GaugeTicks} ticks - the ticks that line the gauge
     * @param {[GaugeValueElement]} valueElements - the element(s) which show the current value on
     * the gauge
     * @param {Vector} centre - the centre of the gauge
     */
    constructor(numbering, ticks, valueElements, centre) {
      super();
      this.numbering = numbering;
      this.ticks = ticks;
      this.valueElements = valueElements;

      // The value being displayed on the gauge
      this.value = 0;
      this.setupCentres(centre);
    }

    /**
     * Sets the current value for the gauge
     * @param {number} value - the value to set
     */
    setValue(value) {
      this.value = value;
    }

    /**
     * Sets the angle of the moving (GaugeValueElements) elements
     * @param {number} angle - the angle to set
     */
    setAngle(angle) {
      for (let element of this.valueElements)
        element.setActiveAngle(angle);
    }

    /**
     * Sets the centre for each component of the gauge
     * @param {Vector} centre - the centre to set
     */
    setupCentres(centre) {
      for (let element of this.valueElements) {
        // Only setup rotational centres for pointer(s)
        if (element instanceof GaugePointer)
          element.setRotationalCentre(TachoBase.Vector.copy(centre))
        else
          element.setCentre(TachoBase.Vector.copy(centre));
      }

      this.ticks.setCentre(TachoBase.Vector.copy(centre));
      this.numbering.setCentre(TachoBase.Vector.copy(centre));
    }

    getValueElements() {
      return this.valueElements;
    }

    getTicks() {
      return this.ticks;
    }

    getNumbering() {
      return this.numbering;
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      for (let element of this.valueElements)
        element.draw(ctx);

      this.numbering.draw(ctx);
      this.ticks.drawTicks(ctx, this.value);
    }
  }

  /**
   * Represents a gauge, similar to Gauge
   */
  class RedlineGauge extends Gauge {
    /**
     * Constructor
     * @param {GaugeNumbers} numbering - the numbers along the gauge
     * @param {GaugeTicks} ticks - the ticks that line the gauge
     * @param {[GaugeValueElement]} valueElements - the element(s) which show the current value on
     * the gauge
     * @param {Vector} centre - the centre of the gauge
     * @param {number} redlineValue - the redline value
     */
    constructor(numbering, ticks, valueElements, centre, redlineValue) {
      super(numbering, ticks, valueElements, centre);
      this.redlineValue = redlineValue;
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      for (let element of this.valueElements)
        element.draw(ctx);

      this.numbering.draw(ctx);
      this.ticks.drawTicks(ctx, this.redlineValue);
    }
  }


  /**
   * Represents a bar gauge displayed as an arc
   */
  class ArcBarGauge extends TachoBase.Drawable {
    constructor(maxValue, radius, arcAngles, styling, centre, invert) {
      super();
      // @type ArcBar
      this.arcBar = new TachoShapes.ArcBar(radius, arcAngles, styling, centre, invert);
      // The current value to display
      this.value = 0;
      // The value at which the bar is full
      this.MAX_VALUE = maxValue;
    }

    /**
     * Sets the value to display
     * @param {number} value - the value to display
     */
    setValue(value) {
      this.value = value;
      this.arcBar.setPercentageAlong(value / this.MAX_VALUE);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      this.arcBar.draw(ctx);
    }
  }



  this.GaugePointer = GaugePointer
  this.SimplePointer = SimplePointer;
  this.AccentPointer = AccentPointer;
  this.GaugeTicks = GaugeTicks;
  this.BiColourGaugeTicks = BiColourGaugeTicks
  this.GaugeNumbers = GaugeNumbers;
  this.GaugeArcs = GaugeArcs;
  this.Gauge = Gauge;
  this.RedlineGauge = RedlineGauge;

  this.ArcBarGauge = ArcBarGauge;

  return this;
}])