angular.module("beamng.apps")
/**
 * Provides shape classes and some misc maths functions
 */
.factory("TachoShapes", ["TachoBase", function(TachoBase) {
  // Mimic inner class behaviour by providing this service to those classes
  let outerThis = this;

  /////////////////////
  // Maths Functions //
  /////////////////////

  /**
   * Returns the coordinates of a point on a circle, whose centre is the origin (0,0)
   * @param {number} radius - the radius of the circle
   * @param {number} angle - the angle along the circle the point is
   * @returns {Vector} the point
   */
  this.getCircularPoint = function(radius, angle) {
    return new TachoBase.Vector(radius * Math.cos(angle), radius * Math.sin(angle));
  }

  /**
   * Returns the coordinates of a point on a circle, whose centre is specified
   * @param {Vector} relCentre - the centre of the circle to rotate around
   * @param {number} radius - the radius of the circle
   * @param {number} angle - the angle along the circle the point is
   * @returns {Vector} the point
   */
  this.getRelCircularPoint = function(relCentre, radius, angle) {
    return TachoBase.Vector.add(relCentre, this.getCircularPoint(radius, angle));
  }



  /**
   * Abstract
   * Represents a drawable shape
   */
  class Shape extends TachoBase.Rotatable {
    /**
     * Constructor
     * @param {BaseStyle} styling - the styling to apply to the shape
     * @param {Vector} [rotationCentre] - the rotation point during rotation. Defaults to (0,0)
     * @param {Vector} [centre] - the centre point of the shape. Defaults to (0,0)
     */
    constructor(styling, centre, rotationCentre) {
      super(centre);
      this.rotationCentre = rotationCentre || new TachoBase.Vector();
      this.styling = styling;
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      ctx.beginPath();
      this.applyRotation(ctx);
      this.drawContinue(ctx);
      this.applyStyling(ctx);
      TachoBase.Canvas.resetTransformationState(ctx);
    }

    /**
     * Applies the styling of this shape to a 2D canvas context
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to apply this
     * shapes styling to
     */
    applyStyling(ctx) {
      this.styling.apply(ctx);
    }

    /**
     * Sets the centre of rotation for this shape
     * @param centre
     */
    setRotationalCentre(centre) {
      this.rotationCentre = centre;
    }

    /**
     * Applies the rotation transformation of a canvas 2D context for drawing the rotated shape
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to rotate
     */
    applyRotation(ctx) {
      ctx.translate(this.rotationCentre.x(), this.rotationCentre.y());
      ctx.rotate(this.angle);
      ctx.translate(-this.rotationCentre.x(), -this.rotationCentre.y());
    }

    /**
     * Draws the arc on a canvas context object. Unlike draw(), this method does not begin nor end
     * the path for the given context object. Before calling this method, beginPath() should be
     * called to ensure correct rendering
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    drawContinue(ctx) {
      console.error("drawContinue() must be implemented by base class")
    }
  }

  /**
   * Represents a drawable arc
   */
  class Arc extends Shape {
    /**
     * Constructor
     * @param {number} radius - the radius of the arc
     * @param {Range} angles - the starting and end angles of the arc
     * @param {BaseStyle} styling - the styling to apply to the arc
     * @param {Vector} [centre] - the centre of the arc. Defaults to (0,0)
     */
    constructor(radius, angles, styling, centre) {
      super(styling, centre);
      this.radius = radius;
      this.angles = angles;
    }

    /**
     * @param {number} startAngle - the starting angle of the arc to set
     */
    setStartAngle(startAngle) { this.angles.setLower(startAngle); }

    /**
     * @param {number} endAngle - the ending angle of the arc to set
     */
    setEndAngle(endAngle) { this.angles.setUpper(endAngle); }

    getStartAngle() {
      return this.angles.lower;
    }

    getEndAngle() {
      return this.angles.upper;
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawContinue(ctx) {
      ctx.arc(this.centre.x(), this.centre.y(), this.radius, this.angles.lower, this.angles.upper);
    }
  }

  /**
   * Represents a drawable circle
   */
  class Circle extends Arc {
    /**
     * Constructor
     * @param {number} radius - the radius of the circle
     * @param {BaseStyle} styling - the styling to apply to the circle
     * @param {Vector} [centre] - the centre of the circle. Defaults to (0,0)
     */
    constructor(radius, styling, centre) {
      super(radius, new TachoBase.Range(0, 2*Math.PI), styling, centre);
    }
  }

  class Wedge extends Arc {
    /**
     * Constructor
     * @param {number} radius - the radius of the wedge
     * @param {Range} angles - the starting and end angles of the wedge
     * @param {BaseStyle} styling - the styling to apply to the wedge
     * @param {Vector} [centre] - the centre of the wedge. Defaults to (0,0)
     */
    constructor(radius, angles, styling, centre) {
      super(radius, angles, styling, centre);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      ctx.beginPath();
      ctx.moveTo(this.centre.x(), this.centre.y());
      super.drawContinue(ctx);
      this.styling.apply(ctx);
    }
  }

  /**
   * Represents a drawable line
   */
  class Line extends Shape {
    /**
     * Constructor
     * @param {Vector} point1 - the start point
     * @param {Vector} point2 - the end point
     * @param {BaseStyle} styling - the styling to apply to the line
     */
    constructor(point1, point2, styling) {
      super(styling);
      this.point1 = point1;
      this.point2 = point2;
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      let oldCentre = this.centre;
      super.setCentre(centre);
      let translation = TachoBase.Vector.sub(this.centre, oldCentre);

      this.translate(translation);
    }

    /**
     * Translates both end points by the given vector
     * @param {Vector} translation the translation vector
     */
    translate(translation) {
      this.point1.add(translation);
      this.point2.add(translation);
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawContinue(ctx) {
      ctx.moveTo(this.point1.x(), this.point1.y());
      ctx.lineTo(this.point2.x(), this.point2.y());
    }
  }

  /**
   * Represents a drawable triangle
   */
  class Triangle extends Line {
    /**
     * Constructor
     * @param {Vector} point1 - the first vertex
     * @param {Vector} point2 - the second vertex
     * @param {Vector} point3 - the third vertex
     * @param {BaseStyle} styling - the styling to apply to the triangle
     */
    constructor(point1, point2, point3, styling) {
      super(point1, point2, styling);
      this.point3 = point3;

      // Calculate and set centre
      this.setCentre(this.calculateCentrePoint());
    }

    /**
     * @Override
     * @inheritDoc
     */
    translate(translation) {
      super.translate(translation);
      this.point3.add(translation);
    }

    /**
     * Constructs and returns an isosceles triangle.
     * @param {number} baseWidth - the width of the base
     * @param {number} height - the height
     * @param {BaseStyle} styling - the styling to apply to the triangle
     * @param {Vector} centre - the point to put the triangle's centre to
     */
    static newIsosceles(baseWidth, height, styling, centre) {
      let p1 = new TachoBase.Vector( baseWidth/2, 0);
      let p2 = new TachoBase.Vector(-baseWidth/2, 0);
      let p3 = new TachoBase.Vector(0, height);

      let triangle = new Triangle(p3, p1, p2, styling);
      triangle.setCentre(centre);
      return triangle;
    }

    /**
     * Calculates the centre point of the triangle
     * @returns {Vector}
     */
    calculateCentrePoint() {
      let sum = new TachoBase.Vector();
      sum.add(this.point1);
      sum.add(this.point2);
      sum.add(this.point3);
      return TachoBase.Vector.div(sum, 3);
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawContinue(ctx) {
      super.drawContinue(ctx);
      ctx.lineTo(this.point3.x(), this.point3.y());
      ctx.lineTo(this.point1.x(), this.point1.y());
    }
  }

  /**
   * Represent a rectangle
   */
  class Rectangle extends Shape {
    /**
     * Constructor
     * @param {number} width - the width of the rectangle
     * @param {number} height - the height of the rectangle
     * @param {BaseStyle} styling - the styling to apply
     * @param {Vector} [centre] - the position of the top left point. Defaults to (0,0)
     * @param {boolean} [isFillCentred] - whether to centre the X position of the rectangle with the
     * fill style. Defaults to false
     */
    constructor(width, height, styling, centre, isFillCentred) {
      super(styling, centre);
      this.width = width;
      this.height = height;
      this.isFillCentred = isFillCentred === true;
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      super.setCentre(centre);

      if (this.isFillCentred)
        this.centre.addX(-this.width/4);
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawContinue(ctx) {
      ctx.rect(this.centre.x(), this.centre.y(), this.width, this.height);
    }
  }

  /**
   * Represents a rectangle with a rounded bottom-right corner
   */
  class RoundCornerRectangleBR extends Shape {
    /**
     * Constructor
     * @param {number} arcRadius - the rounded part's arc radius
     * @param {Range} arcAngles - the rounded part's arc angle range
     * @param {Vector} arcCentre - the rounded part's arc centre point
     * @param {number} leftX - the leftmost point on the x dimension
     * @param {BaseStyle} styling - the style to apply to this shape
     */
    constructor(arcRadius, arcAngles, arcCentre, leftX, styling) {
      super(styling);
      this.arc = new Arc(arcRadius, arcAngles, styling, arcCentre);
      this.leftX = leftX;
    }

    drawContinue(ctx) {
      this.arc.drawContinue(ctx);
      ctx.lineTo(this.leftX, outerThis.getRelCircularPoint(this.arc.centre, this.arc.radius, this.arc.getEndAngle()).y());
      ctx.lineTo(this.leftX, outerThis.getRelCircularPoint(this.arc.centre, this.arc.radius, this.arc.getStartAngle()).y());
    }
  }


  /**
   * Represents a completion bar in an arc shape
   */
  class ArcBar extends TachoBase.Drawable {
    /**
     * Constructor
     * By default, the bar is 100% complete
     * @param {number} radius - the radius of the arc
     * @param {Range} angles - the starting and end angles of the arc
     * @param {BaseStyle} styling - the styling to apply to the arc
     * @param {Vector} [centre] - the centre of the arc. Defaults to (0,0)
     * @param {boolean} [invert] - whether to invert {@param angles} such that the end angle becomes
     * the representation for 0
     */
    constructor(radius, angles, styling, centre, invert) {
      super();
      this.invert = invert;
      this.bgArc = new Arc(radius, angles, new TachoBase.Style("rgba(255, 255, 255, 0.25)", styling.lineWidth, false), centre);
      this.mainArc = new Arc(radius, angles.clone(), styling, centre);
    }

    /**
     * @param {number} perc - the percentage full to set the bar to be (between [0,1])
     */
    setPercentageAlong(perc) {
      if (this.invert) {
        let angle = this.bgArc.angles.getLinearRatio(1 - perc);
        this.mainArc.setStartAngle(angle);
      } else {
        let angle = this.bgArc.angles.getLinearRatio(perc);
        this.mainArc.setEndAngle(angle);
      }
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      ctx.lineCap = "round";
      this.bgArc.draw(ctx);
      this.mainArc.draw(ctx);
      ctx.lineCap = "butt";
    }
  }


  /**
   * Represents a drawable arrow to a canvas. By default (rotation of 0) the arrow points right and
   * the base dimensions are 10x10 px
   */
  class Arrow extends Shape {
    static X_COORDS = {
      BASE: 0,
      ARR_START: 5,
      ARR_END: 10
    }

    static Y_COORDS = {
      BASE_BOT: 2.5,
      ARR_START: 5
    }

    /**
     * Constructor
     * @param {number} scale - the scale factor of the arrow
     * @param {Style} styling - the styling to apply to the arrow
     * @param {Vector} [centre] - the centre point of the arrow (defaults to (0,0))
     */
    constructor(scale, styling, centre) {
      super(styling, centre);
      this.scale = scale;
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawContinue(ctx) {
      // Begins bottom right of rectangular base, heading counter-clockwise

      // Bottom half
      ctx.translate(this.centre.x(), this.centre.y());
      ctx.moveTo(Arrow.X_COORDS.BASE      * this.scale, Arrow.Y_COORDS.BASE_BOT  * this.scale);
      ctx.lineTo(Arrow.X_COORDS.ARR_START * this.scale, Arrow.Y_COORDS.BASE_BOT  * this.scale);
      ctx.lineTo(Arrow.X_COORDS.ARR_START * this.scale, Arrow.Y_COORDS.ARR_START * this.scale);
      ctx.lineTo(Arrow.X_COORDS.ARR_END   * this.scale,                                     0);
      // Top half
      ctx.lineTo(Arrow.X_COORDS.ARR_START * this.scale, -Arrow.Y_COORDS.ARR_START * this.scale);
      ctx.lineTo(Arrow.X_COORDS.ARR_START * this.scale, -Arrow.Y_COORDS.BASE_BOT  * this.scale);
      ctx.lineTo(Arrow.X_COORDS.BASE      * this.scale, -Arrow.Y_COORDS.BASE_BOT  * this.scale);
      ctx.lineTo(Arrow.X_COORDS.BASE      * this.scale,  Arrow.Y_COORDS.BASE_BOT  * this.scale);
    }
  }


  // class PBrake extends Shape {
  //   constructor(styling, centre) {
  //     super(styling, centre);
  //   }
  //
  //   static getPath() {
  //     return new Path2D("m74.980469 362.039062c-48.351563-48.351562-74.980469-112.640624-74.980469-181.019531 0-68.378906 26.628906-132.667969 74.980469-181.019531l28.285156 28.285156c-40.796875 40.796875-63.265625 95.039063-63.265625 152.734375 0 57.695313 22.46875 111.9375 63.265625 152.734375zm437.019531-181.019531c0-68.378906-26.628906-132.667969-74.980469-181.019531l-28.285156 28.285156c40.796875 40.796875 63.265625 95.039063 63.265625 152.734375 0 57.695313-22.46875 111.9375-63.265625 152.734375l28.285156 28.285156c48.351563-48.351562 74.980469-112.640624 74.980469-181.019531zm-80 1.046875c0-97.046875-78.953125-176-176-176s-176 78.953125-176 176 78.953125 176 176 176 176-78.953125 176-176zm-40 0c0 74.988282-61.011719 136-136 136s-136-61.011718-136-136c0-74.992187 61.011719-136 136-136s136 61.007813 136 136zm-116 56h-40v40h40zm0-152h-40v121h40zm0 0");
  //   }
  //
  //   /**
  //    * @Override
  //    * @inheritDoc
  //    */
  //   drawContinue(ctx) {
  //     ctx.stroke(PBrake.getPath());
  //   }
  // }


  this.Shape = Shape;
  this.Arc = Arc;
  this.Circle = Circle;
  this.Wedge = Wedge;
  this.Line = Line;
  this.Triangle = Triangle;
  this.Rectangle = Rectangle;
  this.RoundCornerRectangleBR = RoundCornerRectangleBR;
  this.ArcBar = ArcBar;
  this.Arrow = Arrow;
  // this.PBrake = PBrake;


  return this;
}])