angular.module("beamng.apps")
/**
 * The base service for the Velocity themed apps
 */
.factory("ModFusionBase", ["TachoBase", "TachoApps", "TachoGauge", "TachoColours", "TachoShapes", function(TachoBase, TachoApps, TachoGauge, TachoColours, TachoShapes) {
  let outerThis = this;


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Shared Components                                                                            //
  // (These are functions which generate components which will be shared between FI and           //
  //  tachometer apps)                                                                            //
  //////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates the pointer for all Velocity apps
   * @param {Vector} centre - the centre point of the dial
   * @param {Number} majorLength - the length of the pointer on the major side
   * @param {Style} colourStyle - the colour style to apply to the pointer
   * @returns {AccentPointer} the AccentPointer instance
   */
  function createPointer(centre, majorLength, colourStyle) {
    return new TachoGauge.SimplePointer(
            TachoShapes.Triangle.newIsosceles(majorLength/4, majorLength, colourStyle, centre),
            new TachoShapes.Circle(0, colourStyle),
            new TachoShapes.Circle(0, colourStyle)
    );
  }

  /**
   * Creates the GaugeTicks instance for the gauge
   * @param {number} numTicks - the number of ticks to generate
   * @param {number} tickSpacing - the spacing between two ticks, in terms of value
   * @param {number} arcWidth - the width of the arc
   * @param {number} minTickVal - the lowest tick's value. Defaults to 0
   * @param {ColourPriority} accentStyle - the dual-colour style to apply to the ticks
   * @param {number} baseRadius - the base radius of the app this tick instance is for
   * @returns {BiColourGaugeTicks} the ticks instance
   */
  function createTicks(numTicks, tickSpacing, arcWidth, minTickVal, accentStyle, baseRadius, gaugeClass) {
    const WIDTH = 2;
    const MAJ_SPACING = 8;

    return new gaugeClass(
        numTicks,
        tickSpacing,
        WIDTH,
        MAJ_SPACING,
        accentStyle,
        new TachoBase.Range(baseRadius - arcWidth, baseRadius),
        outerThis.GAUGE_RANGE,
        true,
        minTickVal
    );
  }

  /**
   * Creates the GaugeNumbers instance for the gauge
   * @param {number} numNumbers - the number of numbers that will fill the major points on the
   * tick gauge
   * @param {number} fontSize - the size of the font for the numbers
   * @param {number} numbersRadius - the radius of the arc that the numbers will be spaced around
   * @param {number} spacing - the spacing between two numbers
   * @param {number} [startNum] - the starting number. Defaults to 0
   * @param {string} textColour - the colour to apply to the numbers
   * @returns {GaugeNumbers} the GaugeNumbers instance
   */
  function createNumbers(numNumbers, fontSize, numbersRadius, spacing, startNum, textColour) {
    return new TachoGauge.GaugeNumbers(
        numNumbers,
        spacing,
        outerThis.GAUGE_RANGE,
        new TachoBase.Font(outerThis.FONT_NAME, fontSize, true),
        numbersRadius,
        textColour,
        startNum
    );
  }

  /**
   * Draws a glass arc
   * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
   * @param {number} radius - the radius of the glass arc
   * @param {ColourScheme} colourScheme - the colour scheme to apply
   * @param {Vector} centre - the centre of the arc
   */
  function drawGlass(ctx, radius, colourScheme, centre) {
    let paddedRange = outerThis.GAUGE_RANGE.clone();
    paddedRange.pad(0.15);

    let glassArc = new TachoShapes.Arc(
        radius,
        paddedRange.invert(),
        TachoBase.Style.noStrokeFill(colourScheme.BACKGROUND),
        centre
    );

    // Clear the area for alpha to work
    ctx.globalCompositeOperation = "destination-out";
    glassArc.draw(ctx);
    // Then draw the actual glass arc
    ctx.globalCompositeOperation = "source-over";
    glassArc.draw(ctx);
  }


  ////////////////
  // Tachometer //
  ////////////////

  const FT_WIDTH = {
    OUTER: 6,
    RPM_TICKS: 24,
    INNER: 30,
    BARS: 6,
  };

  const FT_RADII = {
    // The radius of the whole UI
    BASE: 150,
    // The radius of the oil and fuel bars (declared below)
    BARS: {}
  };
  // The radius of the outermost circle/stroke
  FT_RADII.OUTER = FT_RADII.BASE - FT_WIDTH.OUTER;
  // The radius of the RPM ticks
  FT_RADII.RPM_TICKS = FT_RADII.OUTER - FT_WIDTH.RPM_TICKS;
  // The radius of the inner/interior part of the gauge
  FT_RADII.INNER = FT_RADII.RPM_TICKS - FT_WIDTH.INNER;
  // The radius of the RPM numbers
  FT_RADII.RPM_NUMS = FT_RADII.RPM_TICKS - FT_WIDTH.RPM_TICKS/2;

  FT_RADII.BARS.FUEL = FT_RADII.OUTER - FT_WIDTH.BARS - 1;
  FT_RADII.BARS.OIL = FT_RADII.BARS.FUEL - 10;


  /**
   * The concrete class representation of this app
   */
  class FusionTachoApp extends TachoApps.TachometerApp {
    // The RPM spacing between two ticks
    static TICK_SPACING = 125;


    constructor(id, document, bngApi, UiUnits) {
      super(id, outerThis.GAUGE_RANGE, outerThis.FONT_NAME, FT_RADII.BASE, document, bngApi, UiUnits);
      this.needsRedraw = false;

      // 130 is the maximal value for temp as seen in BeamNG's Tacho2 app
      this.oilTempBar = this.createArcBar(130, FT_RADII.BARS.OIL , "#ff3333");
      this.fuelBar    = this.createArcBar(  1, FT_RADII.BARS.FUEL, "#0073ff");
    }

    /**
     * @Override
     * @inheritDoc
     */
    createGauge() {
      this.gauge = new TachoGauge.RedlineGauge(
          createNumbers(
              this.getNumRPMThousands(),
              18,
              FT_RADII.RPM_NUMS,
              1,
              0,
              this.colourScheme.TEXT.secondary,
          ),
          createTicks(
              // +1 to include 0
              Math.floor(this.engine.getRedLineRPM() / FusionTachoApp.TICK_SPACING) + 1,
              FusionTachoApp.TICK_SPACING,
              FT_WIDTH.RPM_TICKS,
              0,
              new TachoColours.ColourPriority(this.colourScheme.ACCENT.secondary, "red"),
              FT_RADII.OUTER,
              TachoGauge.BiColourGaugeTicks
          ),
          [
            createPointer(
                this.CENTRE,
                FT_RADII.OUTER,
                new TachoBase.Style(this.colourScheme.TEXT.secondary, 1, false)
            )
          ],
          this.CENTRE,
          this.engine.getMaxRPM()
      );
    }

    /**
     * @Override
     * @inheritDoc
     *
     */
    drawBackgroundCanvas() {
      super.drawBackgroundCanvas();
      let ctx = this.BG_CANVAS.get2DContext();

      let paddedRange = outerThis.GAUGE_RANGE.clone();
      paddedRange.pad(0.15);

      // Creates the various circles for the background
      new TachoShapes.Circle(this.radius, outerThis.BG_COLOURS.OUTER, this.CENTRE)
          .draw(ctx);
      new TachoShapes.Circle(FT_RADII.OUTER, outerThis.BG_COLOURS.TICKS, this.CENTRE)
          .draw(ctx);
      new TachoShapes.Arc(FT_RADII.RPM_TICKS, paddedRange, outerThis.BG_COLOURS.NUMS, this.CENTRE)
          .draw(ctx);

      drawGlass(ctx, FT_RADII.RPM_TICKS, this.colourScheme, this.CENTRE);
    }

    /**
     * @Override
     * @inheritDoc
     * Draws the RPM gauge and the interior
     */
    draw() {
      super.draw();
      this.drawRPM();
      this.drawInterior();
    }

    /**
     * Draws the interior of the tachometer. This includes the background as well as the speed +
     * unit and the currently selected gear
     */
    drawInterior() {
      this.drawInteriorBackground();

      let convertedSpeed = this.UiUnits.speed(this.speed, this.unit);
      this.drawSpeed(Math.floor(convertedSpeed.val), convertedSpeed.unit);

      this.drawParkingBrake();
      this.drawGear();
      this.drawBars();
      this.drawIndicators();
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawIndicator(isLeft, isOn) {
      const MIDDLE_POS = new TachoBase.Vector(this.radius - 30, this.radius + 20);
      const SPACING = 3;

      // If left, rotate around 180deg and place there
      let padding = isLeft ? SPACING : -SPACING;
      let angle = isLeft ? Math.PI : 0;
      let colour = isOn ? "#00ff00" : "black";

      let arrowPos = TachoBase.Vector.sub(MIDDLE_POS, new TachoBase.Vector(padding, 0));

      let arrow = new TachoShapes.Arrow(2, new TachoBase.Style(colour, 0, true), arrowPos);
      arrow.setRotationalCentre(arrowPos);
      arrow.setAngle(angle);
      arrow.draw(this.CANVAS.get2DContext());
    }

    /**
     * Draws the background for the interior part of the gauge
     */
    drawInteriorBackground() {
      let ctx = this.CANVAS.get2DContext();


      // Apply lighting glow
      if (this.lightsState > 0) {
        // Green for low beam, blue for high beam
        let colour = this.lightsState === 1 ? "green" : "rgb(0, 100, 255)";
        ctx.filter = "drop-shadow(-4px -4px 4px " + colour + ") drop-shadow(4px 4px 4px " + colour + ")";
      }

      // Draw the top semi-circle
      new TachoShapes.Circle(FT_RADII.INNER, new TachoBase.Style("#111111", 0, true), this.CENTRE)
          .draw(ctx);

      // After the semi-circle has been draw, apply the glow and reset to prevent it lingering
      ctx.fill();
      TachoApps.BaseApp.resetFilter(ctx);

      // Draw the bottom semi-circle
      new TachoShapes.Arc(
          FT_RADII.INNER,
          new TachoBase.Range(Math.PI, 2*Math.PI),
          new TachoBase.GradientStyle(new TachoBase.LinearGradient(
              new TachoBase.Vector(0, 0),
              new TachoBase.Vector(this.radius*2, this.radius*2),
              ["#555555", "#222222", "black"]
          )),
          this.CENTRE
      ).draw(ctx);

      this.drawInteriorStroke();
    }

    /**
     * Draws the stroke that outlines the interior part
     */
    drawInteriorStroke() {
      let ctx = this.CANVAS.get2DContext();

      // Draw the inner arc that mirrors the angle of the RPM's arc
      new TachoShapes.Arc(
          FT_RADII.INNER,
          outerThis.GAUGE_RANGE,
          new TachoBase.Style(this.colourScheme.ACCENT.secondary, 5, false),
          this.CENTRE
      ).draw(ctx);

      // Draw the remaining part to finish the circle to add an element of 3D
      new TachoShapes.Arc(
          FT_RADII.INNER,
          outerThis.GAUGE_RANGE.invert(),
          new TachoBase.Style("black", 5, false),
          this.CENTRE
      ).draw(ctx);
    }

    /**
     * Draws the speedometer and its units
     * @param {number} speed - the speed of the vehicle
     * @param {string} unit - the unit {@param speed} is in
     */
    drawSpeed(speed, unit) {
      // The y coordinate for the piece of text to align to
      const Y_COORD = this.radius - 10;

      this.drawSpeedText(speed.toString(), Y_COORD);
      this.drawSpeedUnit(unit, Y_COORD);
    }

    /**
     * Draws the current speed
     * @param {string} speed - the current speed
     * @param {number} yCoord - the y coordinate of where the text will be placed
     */
    drawSpeedText(speed, yCoord) {
      // Wrap around to 0 again
      if (parseInt(speed) > 999)
        speed = (parseInt(speed) % 1000).toString();

      // Add in extra zeros to ensure the number if 3 digits long
      speed = "0".repeat(3-speed.length) + speed;


      const FONT_SIZE = 50;
      let ctx = this.CANVAS.get2DContext();
      let pos = new TachoBase.Vector(this.radius - 60, yCoord);


      // Apply a gradient to the text beforehand
      new TachoBase.LinearGradient(
          new TachoBase.Vector(this.radius, 0),
          new TachoBase.Vector(this.radius, yCoord),
          ["white", this.colourScheme.TEXT.primary]
      ).apply(ctx);

      new TachoBase.Text(
          new TachoBase.Font(this.FONT_NAME, FONT_SIZE, false),
          speed.toString(),
          pos
      ).draw(ctx);
    }

    /**
     * Draws the units that the speed is being displayed in
     * @param {string} unit - the unit
     * @param {number} yCoord - the y coordinate of where the text will be placed
     */
    drawSpeedUnit(unit, yCoord) {
      let ctx = this.CANVAS.get2DContext();
      let pos = new TachoBase.Vector(this.radius + 66, yCoord);
      this.drawPrimaryText(ctx, unit, 12, pos);
    }

    /**
     * Draws the gear
     */
    drawGear() {
      const FONT_SIZE = 60;

      let ctx = this.CANVAS.get2DContext();
      let pos = new TachoBase.Vector(this.radius, this.radius + FONT_SIZE);

      // Apply a gradient to the text beforehand
      new TachoBase.LinearGradient(pos, TachoBase.Vector.sub(pos, FONT_SIZE/2), ["white", "black"])
          .apply(ctx);

      // Apply glow
      if (this.engine.shouldShiftUp())
        ctx.filter = ("drop-shadow(0 0 8px " + this.colourScheme.ACCENT.secondary + ")").repeat(2);

      new TachoBase.Text(
          new TachoBase.Font(this.FONT_NAME, FONT_SIZE, false),
          this.gearbox.getGearName(),
          pos
      ).draw(ctx);

      TachoApps.BaseApp.resetFilter(ctx);
    }

    /**
     * Draws the oil and fuel bars
     */
    drawBars() {
      this.fuelBar.setValue(this.fuel);
      this.fuelBar.draw(this.CANVAS.get2DContext());

      this.oilTempBar.setValue(this.engine.getTemperature());
      this.oilTempBar.draw(this.CANVAS.get2DContext());
    }

    /**
     * Draws the parking brake
     */
    drawParkingBrake() {
      let colour = this.pBrakeEngaged ? "orange" : "rgb(24, 24, 24)";
      let ctx = this.CANVAS.get2DContext();

      let pos = new TachoBase.Vector(this.radius - 24, this.radius + 56);

      // Apply glow
      if (this.pBrakeEngaged)
        ctx.filter = "drop-shadow(0 0 8px " + colour + ")";

      new TachoBase.Text(new TachoBase.Font(this.FONT_NAME, 15, true), "PARK", pos, colour)
          .draw(ctx);
      TachoApps.BaseApp.resetFilter(ctx);
    }

    /**
     * Creates an arc bar in this theme's style
     * @param {number} maxValue - the maximal value
     * @param {number} radius - the arc's radius
     * @param {string} colour - the arc's colour
     * @returns {ArcBarGauge} the ArcBar created
     */
    createArcBar(maxValue, radius, colour) {
      return new TachoGauge.ArcBarGauge(
          maxValue,
          radius,
          FusionTachoApp.getBarRange(),
          new TachoBase.Style(colour, FT_WIDTH.BARS, false),
          this.CENTRE,
          true
      );
    }

    /**
     * @returns {Range} the bar range for the oil and fuel bars
     */
    static getBarRange() {
      return new TachoBase.Range(1.85*Math.PI, 2.45*Math.PI);
    }
  }


  //////////////////////
  // Forced Induction //
  //////////////////////

  const FFI_WIDTH = {
    BOOST_NUMBERS: 15,
    OUTER: 3
  }

  const FFI_RADII = {
    BASE: 75,
    DIAL_RADIUS: 10,
    TICKS: 60,
    BOOST_NUMBERS: 50,
    INNER: 40
  };
  FFI_RADII.OUTER = FFI_RADII.BASE - FFI_WIDTH.OUTER;

  /**
   * The concrete class representation of this app
   */
  class FusionBoostApp extends TachoApps.BoostApp {
    // The number of ticks between two numbers on the gauge
    static NUM_TICKS_PER_GAP = 3;
    // The number of numbers on the gauge to display
    static NUM_NUMBERS = 5;

    constructor(id, document, bngApi, UiUnits) {
      super(id, outerThis.GAUGE_RANGE, outerThis.FONT_NAME, FFI_RADII.BASE, document, bngApi, UiUnits);
    }

    /**
     * @Override
     * @inheritDoc
     */
    createGauge() {
      const NUM_TICKS = FusionBoostApp.NUM_TICKS_PER_GAP * (FusionBoostApp.NUM_NUMBERS-1)
          + FusionBoostApp.NUM_NUMBERS;
      const UNIT_MAX_BOOST = this.UiUnits.pressure(this.maxBoost, this.unit).val;
      const FLOOR_MAX_BOOST = Math.floor(UNIT_MAX_BOOST + 1);
      // Increment between numbers on the gauge
      const NUMBER_INC = FLOOR_MAX_BOOST / Math.floor(FusionBoostApp.NUM_NUMBERS / 2);
      const TICK_SPACING = NUMBER_INC / (FusionBoostApp.NUM_TICKS_PER_GAP + 1);

      const FONT_SIZE = 10 - FLOOR_MAX_BOOST.toString().length;


      this.gauge = new TachoGauge.Gauge(
          createNumbers(
              FusionBoostApp.NUM_NUMBERS,
              FONT_SIZE,
              FFI_RADII.BOOST_NUMBERS,
              NUMBER_INC,
              -FLOOR_MAX_BOOST,
              this.colourScheme.TEXT.secondary
          ),
          createTicks(
              NUM_TICKS,
              TICK_SPACING,
              FFI_WIDTH.BOOST_NUMBERS,
              -UNIT_MAX_BOOST,
              this.colourScheme.ACCENT,
              FFI_RADII.OUTER,
              TachoGauge.GaugeTicks
          ),
          [
            createPointer(
                this.CENTRE,
                FFI_RADII.OUTER,
                new TachoBase.Style(this.colourScheme.TEXT.secondary, 1, false)
            )
          ],
          this.CENTRE
      );
    }

    /**
     * Draws/updates the background canvas
     */
    drawBackgroundCanvas() {
      super.drawBackgroundCanvas();

      let ctx = this.BG_CANVAS.get2DContext();

      new TachoShapes.Circle(FFI_RADII.BASE, outerThis.BG_COLOURS.OUTER, this.CENTRE)
          .draw(ctx);
      new TachoShapes.Circle(FFI_RADII.OUTER, outerThis.BG_COLOURS.TICKS, this.CENTRE)
          .draw(ctx);
      new TachoShapes.Circle(FFI_RADII.TICKS, outerThis.BG_COLOURS.NUMS, this.CENTRE)
          .draw(ctx);

      drawGlass(ctx, FFI_RADII.TICKS, this.colourScheme, this.CENTRE);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw() {
      super.draw();
      this.drawGauge();
      this.drawInteriorBackground();
      this.drawInterior();
    }

    drawGauge() {
      let ctx = this.CANVAS.get2DContext();

      let boost = this.getUnitisedBoostValue();

      // Normalise from range [-1,1] to [0,1]
      let normBoost = (this.currBoost/this.maxBoost + 1)/2;
      let currBoostAngle = this.BOOST_ANGLE_RANGE.getLinearRatio(normBoost);

      this.gauge.setAngle(currBoostAngle);
      this.gauge.setValue(boost);
      this.gauge.draw(ctx);
    }

    /**
     * Draws the background for the interior part of the gauge
     */
    drawInteriorBackground() {
      let ctx = this.CANVAS.get2DContext();

      // Draw the top semi-circle
      new TachoShapes.Circle(FFI_RADII.INNER, new TachoBase.Style("#111111", 0, true), this.CENTRE)
          .draw(ctx);

      // Draw the bottom semi-circle
      new TachoShapes.Arc(
          FFI_RADII.INNER,
          new TachoBase.Range(Math.PI, 2*Math.PI),
          new TachoBase.GradientStyle(new TachoBase.LinearGradient(
              new TachoBase.Vector(0, 0),
              new TachoBase.Vector(this.radius*2, this.radius*2),
              ["#555555", "#222222", "black"]
          )),
          this.CENTRE
      ).draw(ctx);

      this.drawInteriorStroke();
    }

    /**
     * Draws the stroke that outlines the interior part
     */
    drawInteriorStroke() {
      let ctx = this.CANVAS.get2DContext();

      // Draw the inner arc that mirrors the angle of the RPM's arc
      new TachoShapes.Arc(
          FFI_RADII.INNER,
          outerThis.GAUGE_RANGE,
          new TachoBase.Style(this.colourScheme.ACCENT.secondary, 3, false),
          this.CENTRE
      ).draw(ctx);

      // Draw the remaining part to finish the circle to add an element of 3D
      new TachoShapes.Arc(
          FFI_RADII.INNER,
          outerThis.GAUGE_RANGE.invert(),
          new TachoBase.Style("black", 5, false),
          this.CENTRE
      ).draw(ctx);
    }

    /**
     * Draws the unit
     */
    drawInterior() {
      this.drawBoostValue();
      this.drawUnit();
    }

    /**
     * Draws the boost value to the center of the gauge
     */
    drawBoostValue() {
      const FONT_SIZE = 20;
      let ctx = this.CANVAS.get2DContext();

      let pos = TachoBase.Vector.sub(this.CENTRE, new TachoBase.Vector(0, 12));
      let value = this.getUnitisedBoostValue().toFixed(1);

      // Apply a gradient to the text beforehand
      new TachoBase.LinearGradient(
          new TachoBase.Vector(this.radius, 0),
          new TachoBase.Vector(this.radius, pos.y()),
          ["white", this.colourScheme.TEXT.primary]
      ).apply(ctx);

      new TachoBase.Text(
          // Decrease font size as the number of digits increase to fit
          new TachoBase.Font(outerThis.FONT_NAME, FONT_SIZE - value.length, true),
          value,
          pos
      ).draw(ctx);
    }

    /**
     * Draws the unit text
     */
    drawUnit() {
      const FONT_SIZE = 16;

      let ctx = this.CANVAS.get2DContext();
      let pos = TachoBase.Vector.add(this.CENTRE, new TachoBase.Vector(0, 16));

      new TachoBase.LinearGradient(pos, TachoBase.Vector.sub(pos, FONT_SIZE), ["white", "black"])
          .apply(ctx);

      new TachoBase.Text(
          new TachoBase.Font(outerThis.FONT_NAME, FONT_SIZE, true),
          this.unit,
          pos
      ).draw(ctx);
    }
  }



  this.FusionTachoApp = FusionTachoApp;
  this.FusionBoostApp = FusionBoostApp;

  // The name of the font to use
  this.FONT_NAME = "OCR A Std";
  // The angle range of the gauges
  this.GAUGE_RANGE = new TachoBase.Range(0.5*Math.PI, 1.8*Math.PI);

  this.BG_COLOURS = {
    OUTER: TachoBase.Style.noStrokeFill("#333333"),
    TICKS: TachoBase.Style.noStrokeFill("#161616"),
     NUMS: TachoBase.Style.noStrokeFill("#080808")
  };


  return this;
}]);