angular.module("beamng.apps")
/**
 * The base service for the Velocity themed apps
 */
.factory("ModVelocityBase", ["TachoBase", "TachoApps", "TachoGauge", "TachoColours", "TachoShapes", function(TachoBase, TachoApps, TachoGauge, TachoColours, TachoShapes) {
  let outerThis = this;


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Shared Components                                                                            //
  // (These are functions which generate components which will be shared between FI and           //
  //  tachometer apps)                                                                            //
  //////////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates the pointer for all Velocity apps
   * @param {Number} dialRadius - the radius of the middle dial
   * @param {Vector} centre - the centre point of the dial
   * @param {Number} majorLength - the length of the pointer on the major side
   * @param {ColourPriority} accentStyle - the dual-colour style to apply to the pointer
   * @returns {AccentPointer} the AccentPointer instance
   */
  function createPointer(dialRadius, centre, majorLength, accentStyle) {
    let primaryStyle   = TachoBase.Style.unitStrokeFill(accentStyle.primary);
    let secondaryStyle = TachoBase.Style.unitStrokeFill(accentStyle.secondary);

    return new TachoGauge.AccentPointer(
        new TachoGauge.SimplePointer(
            TachoShapes.Triangle.newIsosceles(majorLength/4, majorLength, primaryStyle, centre),
            TachoShapes.Triangle.newIsosceles(majorLength/4, majorLength/2, primaryStyle, centre),
            null
        ),
        new TachoGauge.SimplePointer(
            TachoShapes.Triangle.newIsosceles(majorLength/12, majorLength, secondaryStyle, centre),
            TachoShapes.Triangle.newIsosceles(majorLength/12, majorLength/2, secondaryStyle, centre),
            null
        ),
        new TachoShapes.Circle(dialRadius, new TachoBase.Style(accentStyle.primary, 5, true), centre)
    )
  }

  /**
   * Creates the GaugeTicks instance for the gauge
   * @param {number} numTicks - the number of ticks to generate
   * @param {number} tickSpacing - the spacing between two ticks, in terms of value
   * @param {number} arcWidth - the width of the arc
   * @param {number} minTickVal - the lowest tick's value. Defaults to 0
   * @param {ColourPriority} accentStyle - the dual-colour style to apply to the ticks
   * @param {number} baseRadius - the base radius of the app this tick instance is for
   * @returns {BiColourGaugeTicks} the ticks instance
   */
  function createTicks(numTicks, tickSpacing, arcWidth, minTickVal, accentStyle, baseRadius) {
    const WIDTH = 2;
    const MAJ_SPACING = 4;

    return new TachoGauge.BiColourGaugeTicks(
        numTicks,
        tickSpacing,
        WIDTH,
        MAJ_SPACING,
        accentStyle,
        // +-2 to not go over bounds
        new TachoBase.Range(baseRadius - arcWidth + 2, baseRadius - 2),
        outerThis.GAUGE_RANGE,
        false,
        minTickVal
    );
  }

  /**
   * Creates the GaugeNumbers instance for the gauge
   * @param {number} numNumbers - the number of numbers that will fill the major points on the
   * tick gauge
   * @param {number} fontSize - the size of the font for the numbers
   * @param {number} numbersRadius - the radius of the arc that the numbers will be spaced around
   * @param {number} spacing - the spacing between two numbers
   * @param {number} [startNum] - the starting number. Defaults to 0
   * @param {string} textColour - the colour to apply to the numbers
   * @returns {GaugeNumbers} the GaugeNumbers instance
   */
  function createNumbers(numNumbers, fontSize, numbersRadius, spacing, startNum, textColour) {
    return new TachoGauge.GaugeNumbers(
        numNumbers,
        spacing,
        outerThis.GAUGE_RANGE,
        new TachoBase.Font(outerThis.FONT_NAME, fontSize, true),
        numbersRadius,
        textColour,
        startNum
    );
  }


  ////////////////
  // Tachometer //
  ////////////////

  const VT_WIDTH = {
    // The current RPM arc that goes around the outside
    RPM_ARC: 30,
    // The interior's stroke between itself and the RPM background
    INNER_STROKE: 5,
    // The width of each bar (fuel and temp gauge)
    BARS: 6
  };
  // The background to the RPM thousand numbers
  VT_WIDTH.RPM_BG = 80 - VT_WIDTH.RPM_ARC;

  const VT_RADII = {
    // The radius of the whole UI
    BASE: 150,
    // The radius of the inner circle/background part
    BG_INNER: 90,
    // The radius of the pointer's dial
    RPM_DIAL_RADIUS: 30
  };
  // The radius of the stroke that separates the RPM numbers' background and the inner
  // background piece
  VT_RADII.RPM_ARC_BG = VT_RADII.BG_INNER + VT_WIDTH.RPM_BG / 2;
  // The radius of the RPM arc that goes around the outside of the gauge
  VT_RADII.RPM_ARC = VT_RADII.BASE - VT_WIDTH.RPM_ARC / 2;
  // The radius of the inner (stroked) part of the RPM arc
  VT_RADII.RPM_INNER = VT_RADII.BASE - VT_WIDTH.RPM_ARC;
  // The radius of the inner background piece
  VT_RADII.MIDDLE = VT_RADII.BG_INNER - VT_WIDTH.INNER_STROKE * 2;
  // The radius the RPM numbers are displayed around
  VT_RADII.RPM_NUMBERS = VT_RADII.RPM_INNER - (VT_RADII.RPM_INNER - VT_RADII.BG_INNER + VT_WIDTH.INNER_STROKE) / 2;
  // The radius of the temperature bar
  VT_RADII.TEMP = VT_RADII.MIDDLE;
  VT_RADII.FUEL = VT_RADII.TEMP - 10;


  /**
   * Creates the GaugeArcs instance for the gauge
   * @param arcRadius
   * @param arcWidth
   * @param redlineAngle
   * @param {string} colour - the colour of the main gauge
   * @returns {GaugeArcs} the GaugeArcs instance
   */
  function createArcs(arcRadius, arcWidth, redlineAngle, colour){
    return new TachoGauge.GaugeArcs(
        new TachoShapes.Arc(
            arcRadius,
            // Will be overridden later, so default to no arc
            new TachoBase.Range(outerThis.GAUGE_RANGE.lower, outerThis.GAUGE_RANGE.lower),
            new TachoBase.Style(colour, arcWidth)
        ),
        new TachoShapes.Arc(
            arcRadius,
            new TachoBase.Range(redlineAngle, outerThis.GAUGE_RANGE.upper),
            new TachoBase.Style("red", arcWidth)
        )
    );
  }


  /**
   * The concrete class representation of this app
   */
  class VelocityTachoApp extends TachoApps.TachometerApp {
    // The width of stroke effects for various components
    static WIDTH = VT_WIDTH;

    /*
     * The centre radius values for various components.
     * It should be noted that the radii values are with respect to an arc that they represent.
     * So, a radius value is not the radius of that element, but the radius of the arc around the
     * centre of the UI it will be drawn as.
     * Any stroke effect adds thickness to a radius
     */
    static RADII = VT_RADII;

    // The RPM spacing between two ticks
    static TICK_SPACING = 250;


    constructor(id, document, bngApi, UiUnits) {
      super(id, outerThis.GAUGE_RANGE, outerThis.FONT_NAME, VelocityTachoApp.RADII.BASE, document, bngApi, UiUnits);
      this.needsRedraw = false;

      // 130 is the maximal value for temp as seen in BeamNG's Tacho2 app
      this.oilTempBar = this.createArcBar(130, VelocityTachoApp.RADII.TEMP, "#ff3333");
      this.fuelBar    = this.createArcBar(  1, VelocityTachoApp.RADII.FUEL, "#0073ff");
    }

    /**
     * @Override
     * @inheritDoc
     */
    createGauge() {
      this.gauge = new TachoGauge.Gauge(
        createNumbers(
            this.getNumRPMThousands(),
            20,
            VelocityTachoApp.RADII.RPM_NUMBERS,
            1,
            0,
            this.colourScheme.TEXT.secondary,
        ),
        createTicks(
            // +1 to include 0
            Math.floor(this.engine.getRedLineRPM() / VelocityTachoApp.TICK_SPACING) + 1,
            VelocityTachoApp.TICK_SPACING,
            VelocityTachoApp.WIDTH.RPM_ARC,
            0,
            this.colourScheme.ACCENT,
            this.radius
        ),
        [
            createPointer(
                VelocityTachoApp.RADII.RPM_DIAL_RADIUS,
                this.CENTRE,
                this.radius - VelocityTachoApp.WIDTH.RPM_ARC,
                this.colourScheme.ACCENT
            ),
            createArcs(
                VelocityTachoApp.RADII.RPM_ARC,
                VelocityTachoApp.WIDTH.RPM_ARC,
                this.getRedlineRPMAngle(),
                this.colourScheme.ACCENT.secondary
            )
        ],
        this.CENTRE
      );
    }

    /**
     * @Override
     * @inheritDoc
     * This includes both backgrounds for the speedometer as well as the divider between the
     * RPM gauge and inner part. This also draws the background parts of the RPM (see
     * setupRPM()) and draws the gear in the centre (see drawGear())
     */
    drawBackgroundCanvas() {
      super.drawBackgroundCanvas();
      let ctx = this.BG_CANVAS.get2DContext();

      // The background wedge for the last quarter not filled by the RPM
      new TachoShapes.Wedge(
          VelocityTachoApp.RADII.BASE, this.RPM_ANGLE_RANGE.invert(),
          TachoBase.Style.noStrokeFill(this.colourScheme.BASE), this.CENTRE
      ).draw(ctx);
      // The inner circle part
      new TachoShapes.Wedge(
          VelocityTachoApp.RADII.MIDDLE,
          this.RPM_ANGLE_RANGE,
          TachoBase.Style.noStrokeFill(this.colourScheme.BACKGROUND), this.CENTRE
      ).draw(ctx);

      this.drawRPMBackground();
      this.drawGear();
      this.drawSpeedBackground();
      this.drawParkingBrake();
    }

    /**
     * Setups and draws the RPM gauge's background canvas objects.
     * This draws the dividing arcs as well as the backgrounds (such as redline RPM) for the RPM.
     * This also draws the inner parts of the RPM pointer as well as generating the RPM ticks
     */
    drawRPMBackground() {
      const STROKE_WIDTH = 10;
      const RADIUS = VelocityTachoApp.RADII.BG_INNER - STROKE_WIDTH/2;

      let ctx = this.BG_CANVAS.get2DContext();
      let redlineRPMAngle = this.getRedlineRPMAngle();

      // Draw interior arc between RPM background and the interior
      new TachoShapes.Arc(
          RADIUS,
          this.RPM_ANGLE_RANGE,
          new TachoBase.Style(this.colourScheme.ACCENT.primary, STROKE_WIDTH),
          this.CENTRE
      ).draw(ctx);
      // The background arc to the RPM numbers
      new TachoShapes.Arc(
          VelocityTachoApp.RADII.RPM_ARC_BG,
          this.RPM_ANGLE_RANGE,
          new TachoBase.Style(this.colourScheme.BASE, VelocityTachoApp.WIDTH.RPM_BG + 1),
          this.CENTRE
      ).draw(ctx);

      // Draw non-redline background RPM
      new TachoShapes.Arc(
          VelocityTachoApp.RADII.RPM_ARC,
          new TachoBase.Range(this.RPM_ANGLE_RANGE.lower, redlineRPMAngle),
          new TachoBase.Style(this.colourScheme.ACCENT.primary, VelocityTachoApp.WIDTH.RPM_ARC),
          this.CENTRE
      ).draw(ctx);
    }

    /**
     * Draws the gear number in the middle of the tachometer
     */
    drawGear() {
      let gearName = this.gearbox.getGearName();
      // Base font size is 40; if there is >1 digits on the gear then decrease to fit
      let fontSize = 40 / (1 + (gearName.length - 1)*0.5);

      let ctx = this.CANVAS.get2DContext();

      // Apply some manual centering to fix imperfections
      this.drawPrimaryText(ctx, gearName, fontSize, TachoBase.Vector.add(this.CENTRE, 1));
      TachoApps.BaseApp.resetFilter(ctx);
    }

    /**
     * Draws the background for the oil and fuel bars
     */
    drawBarBackground() {
      let radii = VelocityTachoApp.RADII;
      const DIFF = (radii.TEMP - radii.FUEL)/2;
      // The padding either side of the bars
      const PADDING = VelocityTachoApp.WIDTH.INNER_STROKE;
      const MID_RADIUS = radii.TEMP - DIFF - PADDING;

      const WIDTH = DIFF + VelocityTachoApp.WIDTH.BARS*2 + PADDING;

      let range = VelocityTachoApp.getBarRange();
      // Pad out each end
      range.pad(0.2);

      new TachoShapes.Arc(MID_RADIUS, range, new TachoBase.Style("black", WIDTH, false), this.CENTRE)
          .draw(this.CANVAS.get2DContext());
    }

    /**
     * Draws the background polygon for the speedometer
     */
    drawSpeedBackground() {
      let ctx = this.BG_CANVAS.get2DContext();
      outerThis.drawRoundRectRight(ctx, this, 170);

      // Apply lighting glow
      if (this.lightsState > 0) {
        // Green for low beam, blue for high beam
        let colour = this.lightsState === 1 ? "green" : "rgb(0, 100, 255)";
        ctx.filter = "drop-shadow(-8px -8px 8px " + colour + ")";
      }

      ctx.fill();
      TachoApps.BaseApp.resetFilter(ctx);
    }

    /**
     * @returns {Range} the angle range for the parking brake's background polygonal shape
     */
    getPBrakePolygonAngles() {
      return new TachoBase.Range(
          this.RPM_ANGLE_RANGE.lower * 0.475,
          this.RPM_ANGLE_RANGE.lower * 0.825
      );
    }

    /**
     * Draws the background polygon of the parking brake text
     */
    drawParkingBrake() {
      let ctx = this.BG_CANVAS.get2DContext();
      ctx.beginPath();

      new TachoShapes.RoundCornerRectangleBR(this.radius, this.getPBrakePolygonAngles(), this.CENTRE, 170)
          .drawContinue(ctx);

      ctx.lineWidth = 1;
      ctx.fillStyle = "black";
      ctx.fill();

      this.drawParkingBrakeText();
    }

    /**
     * Draws the parking brake text, coloured to indicate whether it is applied or not
     */
    drawParkingBrakeText() {
      let colour = this.pBrakeEngaged ? "orange" : "rgb(24, 24, 24)";

      let centrePos = TachoShapes.getRelCircularPoint(this.CENTRE, VelocityTachoApp.RADII.RPM_ARC, this.getPBrakePolygonAngles().getLinearRatio(0.5));
      // Manual tuning to make it look more central
      centrePos.addY(8);
      centrePos.addX(-20);

      let ctx = this.BG_CANVAS.get2DContext();

      // Apply glow
      if (this.pBrakeEngaged)
        ctx.filter = "drop-shadow(0 0 8px " + colour + ")";

      new TachoBase.Text(new TachoBase.Font(this.FONT_NAME, 15, true), "PARK", centrePos, colour)
          .draw(ctx);
      TachoApps.BaseApp.resetFilter(ctx);
    }

    /**
     * @Override
     * @inheritDoc
     * Draws the RPM gauge and the interior
     */
    draw() {
      super.draw();
      this.drawBars()
      this.drawRPM();
      this.drawInterior();
    }

    /**
     * Draws the interior parts of the tachometer. This includes the speedometer
     */
    drawInterior() {
      // Re-draw background canvas if a redraw is required
      if (this.needsRedraw) {
        this.drawBackgroundCanvas();
        this.needsRedraw = false;
      }

      this.drawGear();

      if (this.engine.shouldShiftUp())
        this.drawShiftPrompt();

      let convertedSpeed = this.UiUnits.speed(this.speed, this.unit);
      this.drawSpeed(Math.floor(convertedSpeed.val), convertedSpeed.unit);

      this.drawIndicators();
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawIndicator(isLeft, isOn) {
      const MIDDLE_POS = new TachoBase.Vector(this.radius + 82, this.radius + 46);
      const SPACING = 36;

      // If left, rotate around 180deg and place there
      let padding = isLeft ? SPACING : -SPACING;
      let angle = isLeft ? Math.PI : 0;
      let colour = isOn ? "#00ff00" : "#333333";

      let arrowPos = TachoBase.Vector.sub(MIDDLE_POS, new TachoBase.Vector(padding, 0));

      let arrow = new TachoShapes.Arrow(2, new TachoBase.Style(colour, 0, true), arrowPos);
      arrow.setRotationalCentre(arrowPos);
      arrow.setAngle(angle);
      arrow.draw(this.CANVAS.get2DContext());
    }

    /**
     * Draws a ring around the gear to indicate to shift up
     */
    drawShiftPrompt() {
      const GEAR_DIAL = this.getGearDial();

      new TachoShapes.Circle(
          GEAR_DIAL.radius,
          new TachoBase.Style(this.colourScheme.TEXT.primary, 6, false),
          this.CENTRE
      ).draw(this.CANVAS.get2DContext());
    }

    /**
     * @returns {Circle} the gear dial
     */
    getGearDial() {
      // Find pointer dial
      for (let element of this.gauge.getValueElements()) {
        if (element instanceof TachoGauge.GaugePointer)
          return element.pointerDial;
      }

      console.log("Pointer dial not found on gauge!");
    }

    /**
     * Draws the temperature and fuel bars in the centre of the UI as well as the bar background
     */
    drawBars() {
      this.drawBarBackground();

      this.fuelBar.setValue(this.fuel);
      this.fuelBar.draw(this.CANVAS.get2DContext());

      this.oilTempBar.setValue(this.engine.getTemperature());
      this.oilTempBar.draw(this.CANVAS.get2DContext());
    }

    /**
     * Creates an arc bar in this theme's style
     * @param {number} maxValue - the maximal value
     * @param {number} radius - the arc's radius
     * @param {string} colour - the arc's colour
     * @returns {ArcBarGauge} the ArcBar created
     */
    createArcBar(maxValue, radius, colour) {
      return new TachoGauge.ArcBarGauge(
          maxValue,
          radius,
          VelocityTachoApp.getBarRange(),
          new TachoBase.Style(colour, VelocityTachoApp.WIDTH.BARS, false),
          this.CENTRE
      );
    }

    /**
     * @returns {Range} the bar range for the oil and fuel bars
     */
    static getBarRange() {
      return new TachoBase.Range(0.65*Math.PI, 1.35*Math.PI);
    }

    /**
     * Draws the speedometer and its units
     * @param {number} speed - the speed of the vehicle
     * @param {string} unit - the unit {@param speed} is in
     */
    drawSpeed(speed, unit) {
      // The spacing between pieces of text and the space between the top of the speed-box and
      // the first piece of text
      const SPACING = 5;
      const SPEED_FONT_SIZE = 32;
      const UNIT_FONT_SIZE = 20;
      let centrePos = TachoShapes.getRelCircularPoint(
          this.CENTRE,
          VelocityTachoApp.RADII.BG_INNER,
          outerThis.roundRectRightAngle().getLinearRatio(0.5)
      );

      let ctx = this.CANVAS.get2DContext();

      // Centre the text
      centrePos.addY(SPACING * 3);
      this.drawPrimaryText(ctx, speed.toString(), SPEED_FONT_SIZE, centrePos);
      // Account for the size of the speed text to ensure no-overlap
      centrePos.addY(SPEED_FONT_SIZE / 2 + SPACING);
      this.drawPrimaryText(ctx, unit, UNIT_FONT_SIZE, centrePos);
    }
  }


  //////////////////////
  // Forced Induction //
  //////////////////////

  const VFI_WIDTH = {
    // The current boost arc that goes around the outside
    BOOST_ARC: 20,
    // The width of each tick
    TICK_WIDTH: 2
  }

  const VFI_RADII = {
    BASE: 75,
    DIAL_RADIUS: 10
  };
  VFI_RADII.BOOST_ARC = VFI_RADII.BASE - VFI_WIDTH.BOOST_ARC/2;
  VFI_RADII.MIDDLE = 40;
  VFI_RADII.BOOST_NUMBERS = (VFI_RADII.BASE - VFI_WIDTH.BOOST_ARC + VFI_RADII.MIDDLE)/2;

  VFI_WIDTH.BOOST_NUMBERS = VFI_RADII.BASE - VFI_WIDTH.BOOST_ARC - VFI_RADII.MIDDLE;

  /**
   * The concrete class representation of this app
   */
  class VelocityBoostApp extends TachoApps.BoostApp {
    // The width of stroke effects for various components
    static WIDTH = VFI_WIDTH;

    // The centre radius values for various components
    // Any stroke effect adds thickness to a radius
    static RADII = VFI_RADII;

    // The number of ticks between two numbers on the gauge
    static NUM_TICKS_PER_GAP = 3;
    // The number of numbers on the gauge to display
    static NUM_NUMBERS = 5;

    constructor(id, document, bngApi, UiUnits) {
      super(id, outerThis.GAUGE_RANGE, outerThis.FONT_NAME, VelocityBoostApp.RADII.BASE, document, bngApi, UiUnits);
    }

    /**
     * @Override
     * @inheritDoc
     */
    createGauge() {
      const NUM_TICKS = VelocityBoostApp.NUM_TICKS_PER_GAP * (VelocityBoostApp.NUM_NUMBERS-1)
          + VelocityBoostApp.NUM_NUMBERS;
      const UNIT_MAX_BOOST = this.UiUnits.pressure(this.maxBoost, this.unit).val;
      const FLOOR_MAX_BOOST = Math.floor(UNIT_MAX_BOOST + 1);
      // Increment between numbers on the gauge
      const NUMBER_INC = FLOOR_MAX_BOOST / Math.floor(VelocityBoostApp.NUM_NUMBERS / 2);
      const TICK_SPACING = NUMBER_INC / (VelocityBoostApp.NUM_TICKS_PER_GAP + 1);

      const FONT_SIZE = 10 - FLOOR_MAX_BOOST.toString().length;


      this.gauge = new TachoGauge.Gauge(
          createNumbers(
              VelocityBoostApp.NUM_NUMBERS,
              FONT_SIZE,
              VelocityBoostApp.RADII.BOOST_NUMBERS,
              NUMBER_INC,
              -FLOOR_MAX_BOOST,
              this.colourScheme.TEXT.secondary
          ),
          createTicks(
              NUM_TICKS,
              TICK_SPACING,
              VelocityBoostApp.WIDTH.BOOST_ARC,
              -UNIT_MAX_BOOST,
              this.colourScheme.ACCENT,
              this.radius
          ),
          [
            createPointer(
                VelocityBoostApp.RADII.DIAL_RADIUS,
                this.CENTRE,
                this.radius - VelocityBoostApp.WIDTH.BOOST_ARC,
                this.colourScheme.ACCENT
            ),
            new TachoGauge.GaugeArcs(
                new TachoShapes.Arc(
                    VelocityBoostApp.RADII.BOOST_ARC,
                    // Will be overridden later, so default to no arc
                    new TachoBase.Range(outerThis.GAUGE_RANGE.lower, outerThis.GAUGE_RANGE.lower),
                    new TachoBase.Style(
                        this.colourScheme.ACCENT.secondary,
                        VelocityBoostApp.WIDTH.BOOST_ARC
                    )
                )
            )
          ],
          this.CENTRE
      );
    }

    /**
     * Draws/updates the background canvas
     */
    drawBackgroundCanvas() {
      super.drawBackgroundCanvas();
      const STROKE_WIDTH = 6;
      const RADIUS = VelocityBoostApp.RADII.MIDDLE - STROKE_WIDTH/2;

      let ctx = this.BG_CANVAS.get2DContext();

      // The inner circle part
      new TachoShapes.Wedge(VelocityBoostApp.RADII.MIDDLE, this.BOOST_ANGLE_RANGE, TachoBase.Style.noStrokeFill(this.colourScheme.BACKGROUND), this.CENTRE)
          .draw(ctx);
      // Draw interior arc between boost arc background and the interior
      new TachoShapes.Arc(RADIUS, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.ACCENT.primary, STROKE_WIDTH), this.CENTRE)
          .draw(ctx);

      // The wedge not covered by the gauge area
      new TachoShapes.Wedge(VelocityBoostApp.RADII.BASE, this.BOOST_ANGLE_RANGE.invert(), TachoBase.Style.noStrokeFill(this.colourScheme.BASE), this.CENTRE)
          .draw(ctx);

      // The background arc to the RPM numbers
      new TachoShapes.Arc(VelocityBoostApp.RADII.BOOST_NUMBERS, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.BASE, VelocityBoostApp.WIDTH.BOOST_NUMBERS+1), this.CENTRE)
          .draw(ctx);
      // Draw background arc to boost gauge
      // Note: -1 to reduce overlap when the other canvas draws over it
      new TachoShapes.Arc(VelocityBoostApp.RADII.BOOST_ARC, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.ACCENT.primary, VelocityBoostApp.WIDTH.BOOST_ARC), this.CENTRE)
          .draw(ctx);

      this.drawUnit();
    }

    /**
     * Draws the boost's unit
     */
    drawUnit() {
      this.drawUnitBackground();
      this.drawUnitText();
    }

    /**
     * Draws the background of the unit area
     */
    drawUnitBackground() {
      let ctx = this.BG_CANVAS.get2DContext();
      outerThis.drawRoundRectRight(ctx, this, 90);
      ctx.fill();
    }

    /**
     * Draws the unit's name in text form
     */
    drawUnitText() {
      const FONT_SIZE = 20;
      const SPACING = 10;

      let ctx = this.BG_CANVAS.get2DContext();

      let centrePos = TachoShapes.getRelCircularPoint(this.CENTRE, VelocityBoostApp.RADII.BOOST_NUMBERS, this.BOOST_ANGLE_RANGE.lower * 0.275);
      centrePos.addY(SPACING);

      new TachoBase.Text(new TachoBase.Font(this.FONT_NAME, FONT_SIZE, true), this.unit, centrePos, this.colourScheme.TEXT.primary)
          .draw(ctx);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw() {
      super.draw();
      let ctx = this.CANVAS.get2DContext();

      let boost = this.UiUnits.pressure(this.currBoost, this.unit).val;

      // Normalise from range [-1,1] to [0,1]
      let normBoost = (this.currBoost/this.maxBoost + 1)/2;
      let currBoostAngle = this.BOOST_ANGLE_RANGE.getLinearRatio(normBoost);

      this.gauge.setAngle(currBoostAngle);
      this.gauge.setValue(boost);
      this.gauge.draw(ctx);
    }
  }


  /**
   * Returns the the angle range for the middle-right rounded rectangle's arc
   * @returns {Range}
   * @see drawRoundRectRight()
   */
  this.roundRectRightAngle = function() {
    let angleRange = this.GAUGE_RANGE;
    return new TachoBase.Range(angleRange.lower * 0.125, angleRange.lower * 0.425);
  }

  /**
   * Draws (but does not complete) a rounded rectangle at the mid-right position of the UI
   * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
   * @param {BaseApp} app - the app drawing this polygon
   * @param {number} leftX - the left-most x coordinate
   */
  this.drawRoundRectRight = function(ctx, app, leftX) {
    ctx.beginPath();
    new TachoShapes.RoundCornerRectangleBR(
        app.radius, this.roundRectRightAngle(), app.CENTRE, leftX, null
    ).drawContinue(ctx);
    ctx.lineWidth = 1;
    ctx.fillStyle = "black";
  }


  this.VelocityTachoApp = VelocityTachoApp;
  this.VelocityBoostApp = VelocityBoostApp;

  // The name of the font to use
  this.FONT_NAME = "Arial Black";
  // The angle range of the gauges
  this.GAUGE_RANGE = new TachoBase.Range(0.5*Math.PI, 2*Math.PI);


  return this;
}]);